"""Contains cross platform ADS extension functions.

:author: David Browne <davidabrowne@gmail.com>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2018-06-11 18:15:53

"""
from typing import Union, Callable, Any, Tuple, Type, Optional, List, Dict
import ctypes
import os
import platform
import socket
import struct
import sys
from contextlib import closing
from functools import wraps

from .utils import platform_is_linux, platform_is_windows, platform_is_freebsd, find_wstring_null_terminator, get_num_of_chars
from .structs import (
    AmsAddr,
    SAmsAddr,
    AdsVersion,
    SAdsVersion,
    SAdsNotificationAttrib,
    SAdsNotificationHeader,
    SAmsNetId,
    SAdsSymbolEntry,
    NotificationAttrib,
    SAdsSumRequest,
)
from .constants import (
    PLCTYPE_STRING,
    PLCTYPE_WSTRING,
    STRING_BUFFER,
    ADSIGRP_SYM_HNDBYNAME,
    PLCTYPE_UDINT,
    ADST_STRING,
    ADST_WSTRING,
    ADSIGRP_SYM_INFOBYNAMEEX,
    ADSIGRP_SYM_VALBYHND,
    ADSIGRP_SYM_RELEASEHND,
    PORT_SYSTEMSERVICE,
    PORT_REMOTE_UDP,
    ADSIGRP_SUMUP_READ,
    ADSIGRP_SUMUP_WRITE,
    DATATYPE_MAP,
    ads_type_to_ctype,
)
from .errorcodes import ERROR_CODES


NOTEFUNC: Optional[Callable] = None

# _adslib can be WinDLL or CDLL depending on OS
_adsDLL: Union["ctypes.WinDLL", "ctypes.CDLL"]

# load dynamic ADS library
if platform_is_windows():  # pragma: no cover, skip Windows test
    dlldir_handle = None
    if sys.version_info >= (3, 8) and "TWINCAT3DIR" in os.environ:
        # Starting with version 3.8, CPython does not consider the PATH environment
        # variable any more when resolving DLL paths. The following works with the default
        # installation of the Beckhoff TwinCAT ADS DLL.
        dll_path = os.environ["TWINCAT3DIR"] + "\\.."
        if platform.architecture()[0] == "64bit":
            dll_path += "\\Common64"
        else:
            dll_path += "\\Common32"
        dlldir_handle = os.add_dll_directory(dll_path)
    try:
        _adsDLL = ctypes.WinDLL("TcAdsDll.dll")  # type: ignore
    finally:
        if dlldir_handle:
            # Do not clobber the load path for other modules
            dlldir_handle.close()
    NOTEFUNC = ctypes.WINFUNCTYPE(  # type: ignore
        ctypes.c_void_p,
        ctypes.POINTER(SAmsAddr),
        ctypes.POINTER(SAdsNotificationHeader),
        ctypes.c_ulong,
    )

elif platform_is_linux():
    adslib_path = None

    for p in sys.path:
        adslib_path = os.path.join(p, "adslib.so")
        if os.path.exists(adslib_path):
            break

    if adslib_path is None:
        raise OSError(f"Failed to locate `adslib.so` library in {sys.path}")

    # For some reason loading on just "adslib.so" always fails, even if it is under
    # sys.path, so manually search for it first
    _adsDLL = ctypes.CDLL(adslib_path)

    NOTEFUNC = ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(SAmsAddr),
        ctypes.POINTER(SAdsNotificationHeader),
        ctypes.c_ulong,
    )
    
elif platform_is_freebsd():
    # try to load local libTcAdsDll.so in favor to global one
    local_adslib = os.path.join(os.path.dirname(__file__), "libTcAdsDll.so")
    if os.path.isfile(local_adslib):
        adslib = local_adslib
    else:
        adslib = "libTcAdsDll.so"

    _adsDLL = ctypes.CDLL(adslib)

    NOTEFUNC = ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(SAmsAddr),
        ctypes.POINTER(SAdsNotificationHeader),
        ctypes.c_ulong,
    )
    
else:  # pragma: no cover, can not test unsupported platform
    raise RuntimeError("Unsupported platform {0}.".format(sys.platform))

callback_store: Dict[Tuple[AmsAddr, int], Callable[[SAmsAddr, SAdsNotificationHeader, int], None]] = dict()


class ADSError(Exception):
    """Error class for errors related to ADS communication."""

    def __init__(
        self, err_code: Optional[int] = None, text: Optional[str] = None
    ) -> None:
        if err_code is not None:
            self.err_code = err_code
            try:
                self.msg = "{} ({}). ".format(ERROR_CODES[self.err_code], self.err_code)
            except KeyError:
                self.msg = "Unknown Error ({0}). ".format(self.err_code)
        else:
            self.msg = ""

        if text is not None:
            self.msg += text

    def __str__(self):
        # type: () -> str
        """Return text representation of the object."""
        return "ADSError: " + self.msg


def router_function(fn: Callable) -> Callable:
    """Raise a runtime error if on Win32 systems.

    Decorator.

    Decorator for functions that interact with the router for the Linux
    implementation of the ADS library.

    Unlike the Windows implementation which uses a separate router daemon,
    the Linux library manages AMS routing in-process. As such, routing must be
    configured programmatically via. the provided API. These endpoints are
    invalid on Win32 systems, so an exception will be raised.

    """

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Callable:
        if platform_is_windows():  # pragma: no cover, skip Windows test
            raise RuntimeError(
                "Router interface is not available on Win32 systems.\n"
                "Configure AMS routes using the TwinCAT router service."
            )
        return fn(*args, **kwargs)

    return wrapper


@router_function
def adsAddRoute(net_id: SAmsNetId, ip_address: str) -> None:
    """Establish a new route in the AMS Router.

    :param pyads.structs.SAmsNetId net_id: net id of routing endpoint
    :param str ip_address: ip address of the routing endpoint

    """
    add_route = _adsDLL.AdsAddRoute
    add_route.restype = ctypes.c_long

    # Convert ip address to bytes (PY3) and get pointer.
    ip_address_p = ctypes.c_char_p(ip_address.encode("utf-8"))

    error_code = add_route(net_id, ip_address_p)

    if error_code:
        raise ADSError(error_code)


def send_raw_udp_message(
    ip_address: str, message: bytes, expected_return_length: int
) -> Tuple[bytes, Tuple[str, int]]:
    """Send a raw UDP message to the PLC and return the response.

    :param str ip_address: ip address of the PLC
    :param bytes message: the message to send to the PLC
    :param int expected_return_length: number of bytes to expect in response
    :rtype: Tuple[bytes, Tuple[str, int]]
    :return: A tuple containing the response and a tuple containing the IP address and port of the
             sending socket
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as sock:  # UDP
        # Listen on any available port for the response from the PLC
        sock.bind(("", 0))

        # Send our data to the PLC
        sock.sendto(message, (ip_address, PORT_REMOTE_UDP))

        # Response should come in in less than .5 seconds, but wait longer to account for slow
        # communications
        sock.settimeout(5)

        # Allow TimeoutError to be raised so user can handle it how they please
        return sock.recvfrom(expected_return_length)


def type_is_string(plc_type: Type) -> bool:
    """Return true if the given class is a string type."""

    # If single char
    if plc_type == PLCTYPE_STRING:
        return True

    # If char array
    if type(plc_type).__name__ == "PyCArrayType":
        if plc_type._type_ == PLCTYPE_STRING:
            return True

    return False


def type_is_wstring(plc_type: Type) -> bool:
    """Return True if the given class is a WSTRING type."""

    # If single char
    if plc_type == PLCTYPE_WSTRING:
        return True

    return False


def get_value_from_ctype_data(read_data: Optional[Any], plc_type: Type) -> Any:
    """Convert ctypes data object to a regular value based on the PLCTYPE_* property.

    Typical usage is:

    .. code:: python

        obj = my_plc_type.from_buffer(my_buffer)
        value = get_value_from_ctype_data(obj, my_plc_type)

    :param read_data: ctypes._CData object
    :param plc_type: pyads.PLCTYPE_* constant (i.e. a ctypes-like type)
    """

    if read_data is None:
        return None

    if type_is_string(plc_type):
        return read_data.value.decode("utf-8")

    if type_is_wstring(plc_type):
        for ix in range(1, len(read_data), 2):
            if (read_data[ix - 1], read_data[ix]) == (0, 0):
                null_idx = ix - 1
                break
        else:
            raise ValueError("No null-terminator found in buffer")
        return bytearray(read_data[:null_idx]).decode("utf-16-le")

    if type(plc_type).__name__ == "PyCArrayType":
        return list(read_data)

    if hasattr(read_data, "value"):
        return read_data.value

    return read_data  # Just return the object itself, don't throw an error


@router_function
def adsAddRouteToPLC(
    sending_net_id: str,
    adding_host_name: str,
    ip_address: str,
    username: str,
    password: str,
    route_name: Optional[str] = None,
    added_net_id: Optional[str] = None,
) -> bool:
    """Embed a new route in the PLC.

    :param pyads.structs.SAmsNetId sending_net_id: sending net id
    :param str adding_host_name: host name (or IP) of the PC being added
    :param str ip_address: ip address of the PLC
    :param str username: username for PLC
    :param str password: password for PLC
    :param str route_name: PLC side name for route, defaults to adding_host_name or the current hostname of this PC
    :param pyads.structs.SAmsNetId added_net_id: net id that is being added to the PLC, defaults to sending_net_id
    :rtype: bool
    :return: True if the provided credentials are correct, False otherwise

    """
    # ALL SENT STRINGS MUST BE NULL TERMINATED
    adding_host_name += "\0"
    added_net_id = added_net_id if added_net_id else sending_net_id
    route_name = route_name + "\0" if route_name else adding_host_name

    username = username + "\0"
    password = password + "\0"

    # The head of the UDP AMS packet containing host routing information
    data_header = struct.pack(
        ">12s", b"\x03\x66\x14\x71\x00\x00\x00\x00\x06\x00\x00\x00"
    )
    data_header += struct.pack(
        ">6B", *map(int, sending_net_id.split("."))
    )  # Sending net ID
    data_header += struct.pack("<H", PORT_SYSTEMSERVICE)  # Internal communication port
    data_header += struct.pack(">2s", b"\x05\x00")  # Write command
    data_header += struct.pack(">4s", b"\x00\x00\x0c\x00")  # Block of unknown
    data_header += struct.pack("<H", len(route_name))  # Length of sender host name
    data_header += route_name.encode("utf-8")  # Sender host name
    data_header += struct.pack(">2s", b"\x07\x00")  # Block of unknown

    actual_data = struct.pack("<H", 6)  # Byte length of AMS ID (always 6)
    actual_data += struct.pack(
        ">6B", *map(int, added_net_id.split("."))
    )  # Net ID being added to the PLC
    actual_data += struct.pack(
        ">2s", b"\x0d\x00"
    )  # Block of unknown (maybe encryption?)
    actual_data += struct.pack("<H", len(username))  # Length of the user name field
    actual_data += username.encode("utf-8")  # PLC Username
    actual_data += struct.pack(">2s", b"\x02\x00")  # Block of unknown
    actual_data += struct.pack("<H", len(password))  # Length of password field
    actual_data += password.encode("utf-8")  # PLC Password
    actual_data += struct.pack(">2s", b"\x05\x00")  # Block of unknown
    actual_data += struct.pack("<H", len(adding_host_name))  # Length of route name
    actual_data += adding_host_name.encode(
        "utf-8"
    )  # Name of route being added to the PLC

    data, addr = send_raw_udp_message(
        ip_address, data_header + actual_data, 32
    )  # PLC response is 32 bytes long

    rcvd_packet_header = data[
        0:12
    ]  # AMS Packet header, seems to define communication type
    # If the last byte in the header is 0x80, then this is a response to our request
    if struct.unpack(">B", rcvd_packet_header[-1:])[0] == 0x80:
        rcvd_PLC_AMS_ID = struct.unpack(">6B", data[12:18])[0]  # AMS ID of PLC
        # Convert to a String AMS ID
        rcvd_AMS_port = struct.unpack(
            "<H", data[18:20]
        )  # Some sort of AMS port? Little endian
        rcvd_command_code = struct.unpack(
            "<2s", data[20:22]
        )  # Command code (should be read) Little endian
        rcvd_protocol_block = data[22:]  # Unknown block of protocol
        rcvd_is_password_correct = rcvd_protocol_block[
            4:7
        ]  # 0x040000 when password was correct, 0x000407 when it was incorrect

        if rcvd_is_password_correct == b"\x04\x00\x00":
            return True
        elif rcvd_is_password_correct == b"\x00\x04\x07":
            return False

    # If we fell through the whole way to the bottom, then we got a weird response
    raise RuntimeError(f"Unexpected response from PLC: {data!r}")


def adsGetNetIdForPLC(ip_address: str) -> str:
    """Get AMS Net ID from IP address.
    
    :param str ip_address: ip address of the PLC
    :rtype: str
    :return: net id of the device at the provided ip address
    
    """
    # The head of the UDP AMS packet containing host routing information
    data_header = struct.pack(
        ">12s", b"\x03\x66\x14\x71\x00\x00\x00\x00\x01\x00\x00\x00"
    )
    data_header += struct.pack(
        ">6B", *[1, 1, 1, 1, 1, 1]
    )  # It doesn't matter what NetID you use here, so just send 1.1.1.1.1.1
    data_header += struct.pack("<H", PORT_SYSTEMSERVICE)  # Internal communication port
    data_header += struct.pack(">4s", b"\x00\x00\x00\x00")  # Block of unknown

    data, addr = send_raw_udp_message(
        ip_address, data_header, 398
    )
    # PLC response should be 395 bytes long, but some PLCs running build 4026+
    # respond with more bytes, so this takes care of that

    rcvd_packet_header = data[
        0:12
    ]  # AMS Packet header, seems to define communication type
    # If the last byte in the header is 0x80, then this is a response to our request
    if struct.unpack(">B", rcvd_packet_header[-1:])[0] == 0x80:
        ams_id_tuple = struct.unpack(">6B", data[12:18])  # AMS ID of PLC
        return ".".join(map(str, ams_id_tuple))

    # If we fell through the whole way to the bottom, then we got a weird response
    raise RuntimeError(f"Unexpected response from PLC: {data!r}")


@router_function
def adsDelRoute(net_id: SAmsNetId) -> None:
    """Remove existing route from the AMS Router.

    :param pyads.structs.SAmsNetId net_id: net id associated with the routing
        entry which is to be removed from the router.

    """
    delete_route = _adsDLL.AdsDelRoute
    delete_route(net_id)


def adsPortOpenEx() -> int:
    """Connect to the TwinCAT message router.

    :rtype: int
    :return: port number

    """
    port_open_ex = _adsDLL.AdsPortOpenEx
    port_open_ex.restype = ctypes.c_long
    port = port_open_ex()

    if port == 0:
        raise RuntimeError("Failed to open port on AMS router.")

    return port


def adsPortCloseEx(port: int) -> None:
    """Close the connection to the TwinCAT message router."""
    port_close_ex = _adsDLL.AdsPortCloseEx
    port_close_ex.restype = ctypes.c_long
    error_code = port_close_ex(port)

    if error_code:
        raise ADSError(error_code)


def adsGetLocalAddressEx(port: int) -> AmsAddr:
    """Return the local AMS-address and the port number.

    :rtype: pyads.structs.AmsAddr
    :return: AMS-address

    """
    get_local_address_ex = _adsDLL.AdsGetLocalAddressEx
    ams_address_struct = SAmsAddr()
    error_code = get_local_address_ex(port, ctypes.pointer(ams_address_struct))

    if error_code:
        raise ADSError(error_code)

    local_ams_address = AmsAddr()
    local_ams_address._ams_addr = ams_address_struct

    return local_ams_address


def adsSetLocalAddress(ams_netid: SAmsNetId) -> None:
    """Change the local NetId.

    :param pyads.structs.SAmsNetId ams_netid: new AmsNetID
    :rtype: None

    """
    set_local_address = _adsDLL.AdsSetLocalAddress
    set_local_address(ams_netid)


def adsSyncReadStateReqEx(port: int, address: AmsAddr) -> Tuple[int, int]:
    """Read the current ADS-state and the machine-state.

    Read the current ADS-state and the machine-state from the
    ADS-server.

    :param port: AMS port
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :rtype: (int, int)
    :return: ads_state, device_state

    """
    sync_read_state_request = _adsDLL.AdsSyncReadStateReqEx

    # C pointer to ams address struct
    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())

    # Current ADS status and corresponding pointer
    ads_state = ctypes.c_uint16()
    ads_state_pointer = ctypes.pointer(ads_state)

    # Current device status and corresponding pointer
    device_state = ctypes.c_uint16()
    device_state_pointer = ctypes.pointer(device_state)

    error_code = sync_read_state_request(
        port, ams_address_pointer, ads_state_pointer, device_state_pointer
    )

    if error_code:
        raise ADSError(error_code)

    return ads_state.value, device_state.value


def adsSyncReadDeviceInfoReqEx(port: int, address: AmsAddr) -> Tuple[str, AdsVersion]:
    """Read the name and the version number of the ADS-server.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :rtype: string, AdsVersion
    :return: device name, version

    """
    sync_read_device_info_request = _adsDLL.AdsSyncReadDeviceInfoReqEx

    # Get pointer to the target AMS address
    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())

    # Create buffer to be filled with device name, get pointer to said buffer
    device_name_buffer = ctypes.create_string_buffer(20)
    device_name_pointer = ctypes.pointer(device_name_buffer)

    # Create ADS Version struct and get pointer.
    ads_version = SAdsVersion()
    ads_version_pointer = ctypes.pointer(ads_version)

    error_code = sync_read_device_info_request(
        port, ams_address_pointer, device_name_pointer, ads_version_pointer
    )

    if error_code:
        raise ADSError(error_code)

    return device_name_buffer.value.decode(), AdsVersion(ads_version)


def adsSyncWriteControlReqEx(
    port: int,
    address: AmsAddr,
    ads_state: int,
    device_state: int,
    data: Any,
    plc_data_type: Type,
) -> None:
    """Change the ADS state and the machine-state of the ADS-server.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param int ads_state: new ADS-state, according to ADSTATE constants
    :param int device_state: new machine-state
    :param data: additional data
    :param int plc_data_type: plc datatype, according to PLCTYPE constants

    """
    sync_write_control_request = _adsDLL.AdsSyncWriteControlReqEx

    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())
    ads_state_c = ctypes.c_ulong(ads_state)
    device_state_c = ctypes.c_ulong(device_state)

    if type_is_string(plc_data_type):
        data = ctypes.c_char_p(data.encode("utf-8"))
        data_pointer = data
        data_length = len(data_pointer.value) + 1  # add 1 byte for null terminator
    else:
        data = plc_data_type(data)
        data_pointer = ctypes.pointer(data)
        data_length = ctypes.sizeof(data)

    error_code = sync_write_control_request(
        port,
        ams_address_pointer,
        ads_state_c,
        device_state_c,
        data_length,
        data_pointer,
    )

    if error_code:
        raise ADSError(error_code)


def adsSyncWriteReqEx(
    port: int,
    address: AmsAddr,
    index_group: int,
    index_offset: int,
    value: Any,
    plc_data_type: Type,
) -> None:
    """Send data synchronous to an ADS-device.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param int index_group: PLC storage area, according to the INDEXGROUP
        constants
    :param int index_offset: PLC storage address
    :param value: value to write to the storage address of the PLC
    :param int plc_data_type: type of the data given to the PLC,
        according to PLCTYPE constants

    """
    sync_write_request = _adsDLL.AdsSyncWriteReqEx

    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())
    index_group_c = ctypes.c_ulong(index_group)
    index_offset_c = ctypes.c_ulong(index_offset)

    if type_is_string(plc_data_type):
        data = ctypes.c_char_p(value.encode("utf-8"))
        data_pointer = data  # type: Union[ctypes.c_char_p, ctypes.c_wchar_p, ctypes.pointer]
        data_length = len(data_pointer.value) + 1  # type: ignore

    elif type_is_wstring(plc_data_type):
        value_bytes = [byte for byte in value.encode("utf-16-le")]
        data_length = len(value_bytes)  # type: ignore
        data = (data_length * ctypes.c_uint8)(*value_bytes)
        data_pointer = ctypes.pointer(data)

    else:
        if type(plc_data_type).__name__ == "PyCArrayType":
            data = plc_data_type(*value)
        elif type(value) is plc_data_type:
            data = value
        else:
            data = plc_data_type(value)

        data_pointer = ctypes.pointer(data)
        data_length = ctypes.sizeof(data)

    error_code = sync_write_request(
        port,
        ams_address_pointer,
        index_group_c,
        index_offset_c,
        data_length,
        data_pointer,
    )

    if error_code:
        raise ADSError(error_code)


def adsSyncReadWriteReqEx2(
    port: int,
    address: AmsAddr,
    index_group: int,
    index_offset: int,
    read_data_type: Optional[Type],
    value: Any,
    write_data_type: Optional[Type],
    return_ctypes: bool = False,
    check_length: bool = True,
) -> Any:
    """Read and write data synchronous from/to an ADS-device.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param int index_group: PLC storage area, according to the INDEXGROUP
        constants
    :param int index_offset: PLC storage address
    :param Type read_data_type: type of the data given to the PLC to respond to,
        according to PLCTYPE constants, or None to not read anything
    :param value: value to write to the storage address of the PLC
    :param Type write_data_type: type of the data given to the PLC, according to
        PLCTYPE constants, or None to not write anything
    :param bool return_ctypes: return ctypes instead of python types if True
        (default: False)
    :param bool check_length: check whether the amount of bytes read matches the size
        of the read data type (default: True)
    :rtype: read_data_type
    :return: value: value read from PLC

    """
    sync_read_write_request = _adsDLL.AdsSyncReadWriteReqEx2

    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())
    index_group_c = ctypes.c_ulong(index_group)
    index_offset_c = ctypes.c_ulong(index_offset)
    read_data: Optional[Any]
    read_data_pointer: Optional[ctypes.pointer]
    response_size: int = 0

    if index_group == ADSIGRP_SUMUP_READ:
        response_size = 4 * len(value)
        for _ in value:
            response_size += _.size
        read_data_buf = bytearray(response_size)
        read_data = (ctypes.c_ubyte * len(read_data_buf)).from_buffer(read_data_buf)
        read_data_pointer = ctypes.pointer(read_data)
        read_length = response_size

    elif index_group == ADSIGRP_SUMUP_WRITE:
        response_size = (
            index_offset * 4
        )  # expect 4 bytes back for every value written (error data)
        read_data_buf = bytearray(response_size)
        read_data = (ctypes.c_ubyte * len(read_data_buf)).from_buffer(read_data_buf)
        read_data_pointer = ctypes.pointer(read_data)
        read_length = response_size

    elif read_data_type is None:
        read_data = None
        read_data_pointer = None
        read_length = 0
    else:
        if type_is_string(read_data_type):
            read_data = (STRING_BUFFER * PLCTYPE_STRING)()
        elif type_is_wstring(read_data_type):
            read_data = (STRING_BUFFER * ctypes.c_uint8)()
        else:
            read_data = read_data_type()

        read_data_pointer = ctypes.pointer(read_data)
        read_length = ctypes.sizeof(read_data)

    bytes_read = ctypes.c_ulong()
    bytes_read_pointer = ctypes.pointer(bytes_read)

    write_data_pointer: Optional[Union[ctypes.c_char_p, ctypes.c_wchar_p, ctypes.pointer]]
    if index_group == ADSIGRP_SUMUP_READ:
        write_data_pointer = ctypes.pointer(value)
        write_length = ctypes.sizeof(value)
    elif index_group == ADSIGRP_SUMUP_WRITE:
        write_data = (ctypes.c_byte * len(value)).from_buffer(value)
        write_data_pointer = ctypes.pointer(write_data)
        write_length = ctypes.sizeof(write_data)
    elif write_data_type is None:
        write_data_pointer = None
        write_length = 0
    elif type_is_string(write_data_type):
        # Get pointer to string
        write_data_pointer = ctypes.c_char_p(value.encode("utf-8"))
        # Add an extra byte to the data length for the null terminator
        write_length = len(value) + 1
    elif type_is_wstring(write_data_type):
        value_bytes = [byte for byte in value.encode("utf-16-le")]
        write_length = len(value_bytes)  # type: ignore
        write_data = (write_length * ctypes.c_uint8)(*value_bytes)
        write_data_pointer = ctypes.pointer(write_data)
    else:
        if type(write_data_type).__name__ == "PyCArrayType":
            write_data = write_data_type(*value)
        elif type(value) is write_data_type:
            write_data = value
        else:
            write_data = write_data_type(value)
        write_data_pointer = ctypes.pointer(write_data)
        write_length = ctypes.sizeof(write_data)

    err_code = sync_read_write_request(
        port,
        ams_address_pointer,
        index_group_c,
        index_offset_c,
        ctypes.c_ulong(read_length),
        read_data_pointer,
        ctypes.c_ulong(write_length),
        write_data_pointer,
        bytes_read_pointer,
    )

    if err_code:
        raise ADSError(err_code)

    if index_group == ADSIGRP_SUMUP_READ or index_group == ADSIGRP_SUMUP_WRITE:
        expected_length = response_size
    else:
        expected_length = (
            read_data.entryLength
            if isinstance(read_data, SAdsSymbolEntry)
            else read_length
        )

    # If we're reading a value of predetermined size (anything but a string or wstring),
    # validate that the correct number of bytes were read
    if (
        check_length
        and not (type_is_string(read_data_type) or type_is_wstring(read_data_type))
        and bytes_read.value != expected_length
    ):
        raise RuntimeError(
            "Insufficient data (expected {0} bytes, {1} were read).".format(
                expected_length, bytes_read.value
            )
        )

    if return_ctypes:
        return read_data

    return get_value_from_ctype_data(read_data, read_data_type)


def adsSyncReadReqEx2(
    port: int,
    address: AmsAddr,
    index_group: int,
    index_offset: int,
    data_type: Type,
    return_ctypes: bool = False,
    check_length: bool = True,
) -> Any:
    """Read data synchronous from an ADS-device.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param int index_group: PLC storage area, according to the INDEXGROUP
        constants
    :param int index_offset: PLC storage address
    :param Type data_type: type of the data given to the PLC, according to
        PLCTYPE constants
    :param bool return_ctypes: return ctypes instead of python types if True
        (default: False)
    :param bool check_length: check whether the amount of bytes read matches the size
        of the read data type (default: True)
    :rtype: data_type
    :return: value: **value**

    """
    sync_read_request = _adsDLL.AdsSyncReadReqEx2

    ams_address_pointer = ctypes.pointer(address.amsAddrStruct())
    index_group_c = ctypes.c_ulong(index_group)
    index_offset_c = ctypes.c_ulong(index_offset)

    if type_is_string(data_type):
        data = (STRING_BUFFER * PLCTYPE_STRING)()
    elif type_is_wstring(data_type):
        data = (STRING_BUFFER * ctypes.c_uint8)()
    else:
        data = data_type()

    data_pointer = ctypes.pointer(data)
    data_length = ctypes.c_ulong(ctypes.sizeof(data))

    bytes_read = ctypes.c_ulong()
    bytes_read_pointer = ctypes.pointer(bytes_read)

    error_code = sync_read_request(
        port,
        ams_address_pointer,
        index_group_c,
        index_offset_c,
        data_length,
        data_pointer,
        bytes_read_pointer,
    )

    if error_code:
        raise ADSError(error_code)

    # If we're reading a value of predetermined size (anything but a string or wstring),
    # validate that the correct number of bytes were read
    if (
        check_length
        and not(type_is_string(data_type) or type_is_wstring(data_type))
        and bytes_read.value != data_length.value
    ):
        raise RuntimeError(
            "Insufficient data (expected {0} bytes, {1} were read).".format(
                data_length.value, bytes_read.value
            )
        )

    if return_ctypes:
        return data

    return get_value_from_ctype_data(data, data_type)


def adsGetHandle(port: int, address: AmsAddr, data_name: str) -> int:
    """Get the handle of the PLC-variable.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param string data_name: data name
    :rtype: int
    :return: handle: PLC-variable handle
    """
    handle = adsSyncReadWriteReqEx2(
        port,
        address,
        ADSIGRP_SYM_HNDBYNAME,
        0x0,
        PLCTYPE_UDINT,
        data_name,
        PLCTYPE_STRING,
    )

    return handle


def adsGetSymbolInfo(port: int, address: AmsAddr, data_name: str) -> SAdsSymbolEntry:
    """Get the symbol information of the PLC-variable.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param string data_name: data name
    :rtype: SAdsSymbolInfo
    :return: symbol_info: PLC Symbol info
    """
    symbol_info = adsSyncReadWriteReqEx2(
        port,
        address,
        ADSIGRP_SYM_INFOBYNAMEEX,
        0x0,
        SAdsSymbolEntry,
        data_name,
        PLCTYPE_STRING,
    )

    return symbol_info


def adsSumReadBytes(
    port: int,
    address: AmsAddr,
    data_symbols: List[Tuple[int, int, int]],
) -> Any:
    """Perform a sum read for multiple variables, returning the bytes

    This version does not do any processing, and will simply return the concatenation
    of the bytes of the target symbols.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param data_symbols: list of tuples like: (index_group, index_offset, size)
    """
    num_requests = len(data_symbols)
    sum_req_array_type = SAdsSumRequest * num_requests
    sum_req_array = sum_req_array_type()

    for i, data_symbol in enumerate(data_symbols):
        idx_group, idx_offset, size = data_symbol
        sum_req_array[i].iGroup = idx_group
        sum_req_array[i].iOffset = idx_offset
        sum_req_array[i].size = size

    return adsSyncReadWriteReqEx2(
        port,
        address,
        ADSIGRP_SUMUP_READ,
        num_requests,
        None,
        sum_req_array,
        None,
        return_ctypes=False,
        check_length=False,
    )


def adsSumRead(
    port: int, address: AmsAddr, data_names: List[str], data_symbols: Dict[str, SAdsSymbolEntry],
    structured_data_names: List[str],
) -> Dict[str, Any]:
    """Perform a sum read to get the value of multiple variables

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param data_names: list of variables names to read
    :param Dict[str, SAdsSymbolEntry] data_symbols: dictionary of ADS Symbol Info
    :param structured_data_names: list of structured variable names
    :return: result: dict of variable names and values
    :rtype: dict[str, Any]

    """
    result: Dict[str, Optional[Any]] = {i: None for i in data_names}

    num_requests = len(data_names)

    symbol_infos = [
        (data_symbols[name].iGroup, data_symbols[name].iOffs,
         data_symbols[name].size) for name in data_names
    ]
    # When a read is split, `data_symbols` will be bigger than `data_names`
    # Therefore we avoid looping over `data_symbols`

    sum_response = adsSumReadBytes(port, address, symbol_infos)

    data_start = 4 * num_requests
    offset = data_start

    for i, data_name in enumerate(data_names):
        error = struct.unpack_from("<I", sum_response, offset=i * 4)[0]
        if error:
            result[data_name] = ERROR_CODES[error]
        else:
            if data_name in structured_data_names:
                value = sum_response[
                    offset: offset + data_symbols[data_name].size]
            elif data_symbols[data_name].dataType == ADST_STRING:
                num_string_chars = get_num_of_chars(data_symbols[data_name].symbol_type)
                # String size in PLC + null-terminator 1 Byte
                element_size = num_string_chars + 1

                if num_string_chars != -1 and data_symbols[data_name].size/element_size != 1:
                    value = []
                    # For each element - Process the the string and add to value array
                    for j in range(0, int(data_symbols[data_name].size/element_size)):
                        current_offset = offset + (element_size * j)
                        null_idx = sum_response[current_offset: offset + data_symbols[data_name].size].index(0)
                        value.append(bytearray(sum_response[current_offset: current_offset + null_idx]).decode("utf-8"))
                else:
                    # find null-terminator 1 Byte
                    null_idx = sum_response[offset: offset + data_symbols[data_name].size].index(0)
                    value = bytearray(sum_response[offset: offset + null_idx]).decode("utf-8")

            elif data_symbols[data_name].dataType == ADST_WSTRING:
                num_string_chars = get_num_of_chars(data_symbols[data_name].symbol_type)
                # WString size in PLC (2 Bytes per char) + null-terminator 2 Byte
                element_size = ((num_string_chars*2) + 2)
                if num_string_chars != -1 and data_symbols[data_name].size/element_size != 1:
                    value = []
                    # For each element - Process the the string and add to value array
                    for j in range(0, int(data_symbols[data_name].size/element_size)):
                        current_offset = offset + (element_size * j)
                        null_idx = find_wstring_null_terminator(sum_response[current_offset: offset + data_symbols[data_name].size])
                        if null_idx is None:
                            raise ValueError("No null-terminator found in buffer")
                        value.append(bytearray(sum_response[current_offset: current_offset + null_idx]).decode("utf-16-le"))
                else:
                    # find null-terminator 2 Bytes
                    a = sum_response[offset: offset + data_symbols[data_name].size]
                    null_idx = find_wstring_null_terminator(a)
                    if null_idx is None:
                        raise ValueError("No null-terminator found in buffer")
                    value = bytearray(sum_response[offset: offset + null_idx]).decode("utf-16-le")
            elif data_symbols[data_name].size > ctypes.sizeof(ads_type_to_ctype[data_symbols[data_name].dataType]):
                value = list(struct.unpack_from(
                    "<" + DATATYPE_MAP[ads_type_to_ctype[data_symbols[data_name].dataType]][-1] * (data_symbols[data_name].size // ctypes.sizeof(ads_type_to_ctype[data_symbols[data_name].dataType])),
                    sum_response,
                    offset=offset,
                ))
            else:
                value = struct.unpack_from(
                    DATATYPE_MAP[ads_type_to_ctype[data_symbols[data_name].dataType]],
                    sum_response,
                    offset=offset,
                )[0]

            result[data_name] = value
        offset += data_symbols[data_name].size

    return result


def adsSumWriteBytes(
    port: int,
    address: AmsAddr,
    num_requests: int,
    buffer: bytes,
) -> List[str]:
    """Perform a sum write of concatenated bytes to multiple symbols.

    :return: List of errors
    """

    sum_response = adsSyncReadWriteReqEx2(
        port,
        address,
        ADSIGRP_SUMUP_WRITE,
        num_requests,
        None,
        buffer,
        None,
        return_ctypes=False,
        check_length=False,
    )

    errors = list(struct.iter_unpack("<I", sum_response))
    error_descriptions = [ERROR_CODES[i[0]] for i in errors]

    return error_descriptions


def adsSumWrite(
    port: int,
    address: AmsAddr,
    data_names_and_values: Dict[str, Any],
    data_symbols: Dict[str, SAdsSymbolEntry],
    structured_data_names: List[str],
) -> Dict[str, str]:
    """Perform a sum write to write the value of multiple ADS variables

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param data_names_and_values: dict of variable names and values to be written
    :type data_names_and_values: dict[str, Any]
    :param data_symbols: list of dictionaries of ADS Symbol Info
    :type data_symbols: dict[str, ADSSymbolInfo]
    :param structured_data_names: list of structured variable names
    :return: result: dict of variable names and error codes
    :rtype: dict[str, ADSError]
    """
    offset = 0
    num_requests = len(data_names_and_values)
    total_request_size = num_requests * 3 * 4  # iGroup, iOffset & size

    for data_name in data_names_and_values.keys():
        total_request_size += data_symbols[data_name].size

    buf = bytearray(total_request_size)

    for data_name in data_names_and_values.keys():
        struct.pack_into("<I", buf, offset, data_symbols[data_name].iGroup)
        struct.pack_into("<I", buf, offset + 4, data_symbols[data_name].iOffs)
        struct.pack_into("<I", buf, offset + 8, data_symbols[data_name].size)
        offset += 12

    for data_name, value in data_names_and_values.items():
        if data_name in structured_data_names:
            buf[offset: offset + data_symbols[data_name].size] = value
        elif data_symbols[data_name].dataType == ADST_STRING:
            num_string_chars = get_num_of_chars(data_symbols[data_name].symbol_type)
            # String size in PLC + null-terminator 1 Byte
            element_size = (num_string_chars + 1)

            if num_string_chars != -1 and data_symbols[data_name].size/element_size != 1:
                for i, element in enumerate(value):
                    current_offset = offset + (element_size * i)
                    buf[current_offset: current_offset + len(element)] = element.encode("utf-8")
            else:
                buf[offset: offset + len(value)] = value.encode("utf-8")
        elif data_symbols[data_name].dataType == ADST_WSTRING:
            num_string_chars = get_num_of_chars(data_symbols[data_name].symbol_type)
            # WString size in PLC (2 Bytes per char) + null-terminator 2 Byte
            element_size = ((num_string_chars*2) + 2)
            if num_string_chars != -1 and data_symbols[data_name].size/element_size != 1:
                for i, element in enumerate(value):
                    current_offset = offset + (element_size * i)
                    buf[current_offset: current_offset + (2 * len(element))] = element.encode("utf-16-le")
            else:
                buf[offset: offset + 2 * len(value)] = value.encode("utf-16-le")
        elif data_symbols[data_name].size > ctypes.sizeof(ads_type_to_ctype[data_symbols[data_name].dataType]):
            struct.pack_into(
                "<" + DATATYPE_MAP[ads_type_to_ctype[data_symbols[data_name].dataType]][-1] * (data_symbols[data_name].size // ctypes.sizeof(ads_type_to_ctype[data_symbols[data_name].dataType])),
                buf,
                offset,
                *value,
            )
        else:
            struct.pack_into(
                DATATYPE_MAP[ads_type_to_ctype[data_symbols[data_name].dataType]],
                buf,
                offset,
                value,
            )
        offset += data_symbols[data_name].size

    error_descriptions = adsSumWriteBytes(
        port,
        address,
        num_requests,
        buf,
    )

    return dict(zip(data_names_and_values.keys(), error_descriptions))


def adsReleaseHandle(port: int, address: AmsAddr, handle: int) -> None:
    """Release the handle of the PLC-variable.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param int handle: handle of PLC-variable to be released
    """
    adsSyncWriteReqEx(port, address, ADSIGRP_SYM_RELEASEHND, 0, handle, PLCTYPE_UDINT)


def adsSyncReadByNameEx(
    port: int,
    address: AmsAddr,
    data_name: str,
    data_type: Type,
    return_ctypes: bool = False,
    handle: int = None,
    check_length: bool = True,
) -> Any:
    """Read data synchronous from an ADS-device from data name.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param string data_name: data name
    :param Type data_type: type of the data given to the PLC, according to
        PLCTYPE constants
    :param bool return_ctypes: return ctypes instead of python types if True
        (default: False)
    :param int handle: PLC-variable handle (default: None)
    :param bool check_length: check whether the amount of bytes read matches the size
        of the read data type (default: True)
    :rtype: data_type
    :return: value: **value**

    """
    if handle is None:
        no_handle = True
        handle = adsGetHandle(port, address, data_name)
    else:
        no_handle = False

    # Read the value of a PLC-variable, via handle
    value = adsSyncReadReqEx2(
        port,
        address,
        ADSIGRP_SYM_VALBYHND,
        handle,
        data_type,
        return_ctypes,
        check_length,
    )

    if no_handle is True:
        adsReleaseHandle(port, address, handle)

    return value


def adsSyncWriteByNameEx(
    port: int,
    address: AmsAddr,
    data_name: str,
    value: Any,
    data_type: Type,
    handle: int = None,
) -> None:
    """Send data synchronous to an ADS-device from data name.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr address: local or remote AmsAddr
    :param string data_name: PLC storage name
    :param value: value to write to the storage address of the PLC
    :param Type data_type: type of the data given to the PLC,
        according to PLCTYPE constants
    :param int handle: PLC-variable handle (default: None)
    """
    if handle is None:
        no_handle = True
        handle = adsGetHandle(port, address, data_name)
    else:
        no_handle = False

    # Write the value of a PLC-variable, via handle
    adsSyncWriteReqEx(port, address, ADSIGRP_SYM_VALBYHND, handle, value, data_type)

    if no_handle is True:
        adsReleaseHandle(port, address, handle)


def adsSyncAddDeviceNotificationReqEx(
    port: int,
    adr: AmsAddr,
    data: Union[str, Tuple[int, int]],
    pNoteAttrib: NotificationAttrib,
    callback: Callable,
    user_handle: Optional[int] = None,
) -> Tuple[int, int]:
    """Add a device notification.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr adr: local or remote AmsAddr
    :param Union[str, Tuple[int, int]] data: PLC storage address by name or index group and offset
    :param pyads.structs.NotificationAttrib pNoteAttrib: notification attributes
    :param callback: Callback function to handle notification
    :param user_handle: User Handle
    :rtype: (int, int)
    :returns: notification handle, user handle

    """
    global callback_store

    if NOTEFUNC is None:
        raise TypeError("Callback function type can't be None")

    adsSyncAddDeviceNotificationReqFct = _adsDLL.AdsSyncAddDeviceNotificationReqEx

    pAmsAddr = ctypes.pointer(adr.amsAddrStruct())
    if isinstance(data, str):
        hnl = adsSyncReadWriteReqEx2(
            port, adr, ADSIGRP_SYM_HNDBYNAME, 0x0, PLCTYPE_UDINT, data, PLCTYPE_STRING
        )

        nIndexGroup = ctypes.c_ulong(ADSIGRP_SYM_VALBYHND)
        nIndexOffset = ctypes.c_ulong(hnl)
    elif isinstance(data, tuple):
        nIndexGroup = ctypes.c_ulong(data[0])
        nIndexOffset = ctypes.c_ulong(data[1])
        hnl = None
    else:
        raise TypeError(
            "Parameter data has the wrong type %s. Allowed types are: str, Tuple[int, int]."
            % (type(data))
        )

    attrib = pNoteAttrib.notificationAttribStruct()
    pNotification = ctypes.c_ulong()

    nHUser = ctypes.c_ulong(0)
    if hnl is not None:
        nHUser = ctypes.c_ulong(hnl)
    if user_handle is not None:
        nHUser = ctypes.c_ulong(user_handle)

    adsSyncAddDeviceNotificationReqFct.argtypes = [
        ctypes.c_ulong,
        ctypes.POINTER(SAmsAddr),
        ctypes.c_ulong,
        ctypes.c_ulong,
        ctypes.POINTER(SAdsNotificationAttrib),
        NOTEFUNC,  # type: ignore
        ctypes.c_ulong,
        ctypes.POINTER(ctypes.c_ulong),
    ]
    adsSyncAddDeviceNotificationReqFct.restype = ctypes.c_long

    # noinspection PyUnusedLocal
    def wrapper(addr: SAmsAddr, notification: SAdsNotificationHeader, user: int) -> Callable[
            [SAdsNotificationHeader, str], None]:
        return callback(notification, data)

    # noinspection PyTypeChecker
    c_callback = NOTEFUNC(wrapper)  # type: ignore
    err_code = adsSyncAddDeviceNotificationReqFct(
        port,
        pAmsAddr,
        nIndexGroup,
        nIndexOffset,
        ctypes.byref(attrib),
        c_callback,
        nHUser,
        ctypes.byref(pNotification),
    )

    if err_code:
        raise ADSError(err_code)
    callback_store[(adr, pNotification.value)] = c_callback
    return pNotification.value, hnl


def adsSyncDelDeviceNotificationReqEx(
    port: int, adr: AmsAddr, notification_handle: int, user_handle: int
) -> None:
    """Remove a device notification.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param pyads.structs.AmsAddr adr: local or remote AmsAddr
    :param int notification_handle: Notification Handle
    :param int user_handle: User Handle

    """
    adsSyncDelDeviceNotificationReqFct = _adsDLL.AdsSyncDelDeviceNotificationReqEx

    pAmsAddr = ctypes.pointer(adr.amsAddrStruct())
    nHNotification = ctypes.c_ulong(notification_handle)
    err_code = adsSyncDelDeviceNotificationReqFct(port, pAmsAddr, nHNotification)
    del callback_store[(adr, notification_handle)]
    if err_code:
        raise ADSError(err_code)

    if user_handle is not None:
        adsSyncWriteReqEx(
            port, adr, ADSIGRP_SYM_RELEASEHND, 0, user_handle, PLCTYPE_UDINT
        )


def adsSyncSetTimeoutEx(port: int, n_ms: int) -> None:
    """Set Timeout.

    :param int port: local AMS port as returned by adsPortOpenEx()
    :param int n_ms: timeout in ms

    """
    adsSyncSetTimeoutFct = _adsDLL.AdsSyncSetTimeoutEx
    cms = ctypes.c_long(n_ms)
    err_code = adsSyncSetTimeoutFct(port, cms)
    if err_code:
        raise ADSError(err_code)



"""Abstract handler module for testserver.

:author: David Browne <davidabrowne@gmail.com>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2016-09-13

"""
import logging
from collections import namedtuple


# Log to stdout by default
logger = logging.getLogger(__name__)
formatter = logging.Formatter("%(levelname)s:%(message)s")
stdout_handler = logging.StreamHandler()
stdout_handler.setLevel(logging.WARN)
stdout_handler.setFormatter(formatter)
logger.addHandler(stdout_handler)
logger.setLevel(logging.WARN)
logger.propagate = False  # "Overwrite" default handler

null_logger = logging.getLogger(__name__ + "_null")
null_logger.addHandler(logging.NullHandler())

# Container for data in the 'AMS/TCP header' component of an AMS packet
AmsTcpHeader = namedtuple("AmsTcpHeader", ("length",))

# Container for data in the 'AMS header' component of an AMS packet
AmsHeader = namedtuple(
    "AmsHeader",
    (
        "target_net_id",
        "target_port",
        "source_net_id",
        "source_port",
        "command_id",
        "state_flags",
        "length",
        "error_code",
        "invoke_id",
        "data",
    ),
)

# Container for the entire AMS/TCP packet
AmsPacket = namedtuple("AmsPacket", ("tcp_header", "ams_header"))

# Container for the data required to construct an AMS response given a request
AmsResponseData = namedtuple("AmsResponseData",
                             ("state_flags", "error_code", "data"))


class AbstractHandler:
    """Abstract Handler class to provide a base class for handling requests."""

    def handle_request(self, request: AmsPacket) -> AmsResponseData:
        """Handle incoming requests.

        :param AmsPacket request: The request data received from the client
        :rtype: AmsResponseData
        :return: Data needed to construct the AMS response packet

        """
        raise not NotImplementedError()  # type: ignore



"""Extended ADS TCP/IP server implementation.

Extended ADS TCP/IP server implementation to allow for functional testing of
the ADS protocol without connection to a physical device.

Consists of a server thread which will listen for connections and delegate
each new connection to a separate client thread, allowing for multiple clients
to connect at once.

Each client connection thread listens for incoming data, and delegates parsing
and response construction to the handler. A handler function is injectable at
server level by specifying the `handler` kwarg in the server constructor.

:author: David Browne <davidabrowne@gmail.com>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2018-06-11 18:15:53

"""
from __future__ import absolute_import
from typing import Any, List, Type, Optional
from types import TracebackType
import atexit
import select
import socket
import struct
import threading

from .handler import AbstractHandler, logger, null_logger, AmsPacket, AmsResponseData, AmsHeader, AmsTcpHeader
from .basic_handler import BasicHandler
from .advanced_handler import AdvancedHandler

ADS_PORT = 0xBF02


class AdsTestServer(threading.Thread):
    """Simple ADS testing server.

    :ivar function handler: Request handler (see `default_handler` for example)
    :ivar str ip_address: Host address for server. Defaults to '127.0.0.1'
    :ivar int port: Host port to listen on, defaults to 48898

    """

    def __init__(
            self,
            handler: "AbstractHandler" = None,
            ip_address: str = "127.0.0.1",
            port: int = ADS_PORT,
            logging: bool = True,
            *args: Any,
            **kwargs: Any,
    ) -> None:
        self.handler = handler or BasicHandler()
        self.ip_address = ip_address
        self.port = port
        self._run = True

        global logger
        logger = logger if logging else null_logger

        # Keep track of all received AMS packets
        self.request_history: List[AmsPacket] = []

        # Initialize TCP/IP socket server
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Set option to allow instant socket reuse after shutdown
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.server.bind((self.ip_address, self.port))

        # Make sure we clean up on exit
        atexit.register(self.close)

        # Daemonize the server thread

        # Container for client connection threads
        self.clients: List[AdsClientConnection] = []

        super(AdsTestServer, self).__init__(*args, **kwargs)
        self.daemon = True

    def __enter__(self) -> "AdsTestServer":
        """Enter context."""
        self.start()
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        """Exit context."""
        self.close()

    def stop(self) -> None:
        """Close client connections and stop main server loop."""
        # Close all client connections
        for client in self.clients:
            client.close()

        self.clients = []

        if self._run:
            logger.info("Stopping server thread.")
            # Stop server loop execution
            self._run = False

        self.server.close()

    def close(self) -> None:
        """Close the server thread."""
        self.stop()

    def run(self) -> None:
        """Listen for incoming connections from clients."""
        self._run = True

        # Start server listening
        self.server.listen(5)

        logger.info(
            "Server listening on {0}:{1}".format(
                self.ip_address or "localhost", self.port
            )
        )

        # Server loop
        while self._run:
            # Check for new connections at server socket
            ready, _, _ = select.select([self.server], [], [], 0.1)

            if ready:
                # Accept connection from client
                try:
                    client, address = self.server.accept()
                except:
                    continue

                logger.info("New connection from {0}:{1}".format(*address))

                # Delegate handling of connection to client thread
                client_thread = AdsClientConnection(
                    handler=self.handler, client=client, address=address,
                    server=self
                )

                client_thread.daemon = True
                client_thread.start()
                self.clients.append(client_thread)


class AdsClientConnection(threading.Thread):
    """Connection thread to an ADS client."""

    def __init__(self, handler: "AbstractHandler", client: socket.socket, address: str, server: AdsTestServer,
                 *args: Any,
                 **kwargs: Any) -> None:
        self.handler = handler
        self.server = server
        self.client = client
        self.client_address = address

        # Make sure we clean up on exit
        atexit.register(self.close)

        # Server loop execution flag
        self._run = True

        super(AdsClientConnection, self).__init__(*args, **kwargs)

    def stop(self) -> None:
        """Stop the client thread."""
        if self._run:
            logger.info(
                "Closing client connection {0}:{1}.".format(
                    *self.client_address)
            )
            self._run = False

        self.join()

    def close(self) -> None:
        """Close the client connection."""
        if self.is_alive():
            self.stop()
        self.client.close()

    def run(self) -> None:
        """Listen for data on client connection and delegate requests."""
        self._run = True

        # Main listening loop
        while self._run:
            ready, _, _ = select.select([self.client], [], [], 0.1)

            if not ready:
                continue

            data, _ = self.client.recvfrom(4096)

            if not data:
                self.client.close()
                self._run = False
                continue

            # Basic data validation
            if len(data) < 38:
                logger.warning(
                    "Malformed packet discarded from {0}:{1}:\n\t{data}".format(
                        *self.client_address, data=data
                    )
                )
                continue

            # Construct AmsPacket tuple containing request data
            request_packet = self.construct_request(data)

            self.server.request_history.append(request_packet)

            # Delegate request handling and get response data
            response = self.handler.handle_request(request_packet)

            if isinstance(response, (AmsResponseData,)):
                # Convert request, response data (tuples) to a valid ADS
                # response (bytes) to return to the client
                response_bytes = self.construct_response(response,
                                                         request_packet)

                self.client.send(response_bytes)

                continue

            logger.error("Request handler failed to return a valid response.")

    @staticmethod
    def construct_response(response_data: AmsResponseData, request: AmsPacket) -> bytes:
        """Construct binary AMS response to return to the client.

        :param AmsResponseData response_data: Data to include in the response
        :param AmsPacket request: The originating request for the response

        """
        # Response gets returned to the source, so flip source and target
        target_net_id = request.ams_header.source_net_id
        target_port = request.ams_header.source_port
        source_net_id = request.ams_header.target_net_id
        source_port = request.ams_header.target_port

        # Command ID and invoke ID should be same as in the request
        command_id = request.ams_header.command_id
        invoke_id = request.ams_header.invoke_id

        # Use state flags as specified in response data
        state_flags = response_data.state_flags

        # Calculate payload length and unpack to binary data
        ams_length = struct.pack("<I", len(response_data.data))

        # Use error code specified in response data
        error_code = response_data.error_code

        data = response_data.data

        # Below we [ab]use `encode` to get a py2/3 compatible binary object
        # (str in py2, or bytes in py3)

        # Concatenate ams header data into single binary object
        ams_header = "".encode("utf-8").join(
            (
                target_net_id,
                target_port,
                source_net_id,
                source_port,
                command_id,
                state_flags,
                ams_length,
                error_code,
                invoke_id,
                data,
            )
        )

        ams_tcp_header = "\x00\x00".encode("utf-8") + struct.pack("<I", len(
            ams_header))

        return ams_tcp_header + ams_header

    @staticmethod
    def construct_request(request_bytes: bytes) -> AmsPacket:
        """Unpack an AMS packet from binary data.

        :param bytes request_bytes: The raw request data
        :rtype AmsPacket:
        :return: AmsPacket with fields populated from the binary data

        """
        data = request_bytes  # Use a shorter name for brevity

        tcp_header = AmsTcpHeader(data[2:6])

        ams_header = AmsHeader(
            # Extract target/source net ID's and ports
            data[6:12],
            data[12:14],
            data[14:20],
            data[20:22],
            # Extract command ID, state flags, and data length
            data[22:24],
            data[24:26],
            data[26:30],
            # Extract error code, invoke ID, and data
            data[30:34],
            data[34:38],
            data[38:],
        )

        return AmsPacket(tcp_header, ams_header)



"""The pyads package.

:author: Stefan Lehmann <stlm@posteo.de>
:maintainer: Filippo Boido <filippo.boido@agileautomation.eu> (Agile Automation Technologies GmbH)
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2018-06-11 18:15:53

"""
from .structs import AmsAddr, NotificationAttrib

from .ads import (
    open_port,
    close_port,
    get_local_address,
    add_route,
    add_route_to_plc,
    delete_route,
    set_local_address,
    set_timeout,
    size_of_structure,
    dict_from_bytes,
    bytes_from_dict,
    StructureDef,
)

from .connection import Connection

from .pyads_ex import ADSError

from .constants import (
    PLCTYPE_BOOL,
    PLCTYPE_BYTE,
    PLCTYPE_DATE,
    PLCTYPE_DINT,
    PLCTYPE_DT,
    PLCTYPE_DWORD,
    PLCTYPE_INT,
    PLCTYPE_LINT,
    PLCTYPE_LREAL,
    PLCTYPE_REAL,
    PLCTYPE_SINT,
    PLCTYPE_STRING,
    PLCTYPE_WSTRING,
    PLCTYPE_TIME,
    PLCTYPE_TOD,
    PLCTYPE_ULINT,
    PLCTYPE_UDINT,
    PLCTYPE_UINT,
    PLCTYPE_USINT,
    PLCTYPE_WORD,
    PLCTYPE_ARR_DINT,
    PLCTYPE_ARR_INT,
    PLCTYPE_ARR_LREAL,
    PLCTYPE_ARR_REAL,
    PLCTYPE_ARR_SHORT,
    PLC_DEFAULT_STRING_SIZE,
    DATATYPE_MAP,
)

from .constants import (
    PORT_EVENTLOGGER,
    PORT_IO,
    PORT_LOGGER,
    PORT_NC,
    PORT_NOCKE,
    PORT_SCOPE,
    PORT_SPECIALTASK1,
    PORT_SPECIALTASK2,
    PORT_SPS1,
    PORT_SPS2,
    PORT_SPS3,
    PORT_SPS4,
    PORT_SYSTEMSERVICE,
    PORT_CAM,
    PORT_TC2PLC1,
    PORT_TC2PLC2,
    PORT_TC2PLC3,
    PORT_TC2PLC4,
    PORT_TC3PLC1,
)

from .constants import (
    INDEXGROUP_MEMORYBYTE,
    INDEXGROUP_MEMORYBIT,
    INDEXGROUP_MEMORYSIZE,
    INDEXGROUP_RETAIN,
    INDEXGROUP_RETAINSIZE,
    INDEXGROUP_DATA,
    INDEXGROUP_DATASIZE,
)

from .constants import (
    ADSSTATE_INVALID,
    ADSSTATE_IDLE,
    ADSSTATE_RESET,
    ADSSTATE_INIT,
    ADSSTATE_START,
    ADSSTATE_RUN,
    ADSSTATE_STOP,
    ADSSTATE_SAVECFG,
    ADSSTATE_LOADCFG,
    ADSSTATE_POWERFAILURE,
    ADSSTATE_POWERGOOD,
    ADSSTATE_ERROR,
    ADSSTATE_SHUTDOWN,
    ADSSTATE_SUSPEND,
    ADSSTATE_RESUME,
    ADSSTATE_CONFIG,
    ADSSTATE_RECONFIG,
)

from .constants import (
    ADSTRANS_NOTRANS,
    ADSTRANS_CLIENTCYCLE,
    ADSTRANS_CLIENT1REQ,
    ADSTRANS_SERVERCYCLE,
    ADSTRANS_SERVERONCHA,
)

from .symbol import AdsSymbol

__version__ = '0.1.1'



"""The testserver package of pyads.

:author: Roberto Roos
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2021-04-09

"""
from .testserver import AdsTestServer
from .basic_handler import BasicHandler
from .advanced_handler import AdvancedHandler, PLCVariable
from .handler import AmsTcpHeader, AmsHeader, AmsPacket, AmsResponseData



"""Basic handler module for testserver.

:author: Stefan Lehmann <stlm@posteo.de>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2017-09-15

"""
from typing import List, Union
import struct

from .handler import AbstractHandler, AmsPacket, AmsResponseData, logger
from pyads import constants


class BasicHandler(AbstractHandler):
    """Basic request handler.

    Basic request handler to print the request data and return some default values.

    """

    def handle_request(self, request: AmsPacket) -> AmsResponseData:
        """Handle incoming requests and send a response."""
        # Extract command id from the request
        command_id_bytes = request.ams_header.command_id
        command_id = struct.unpack("<H", command_id_bytes)[0]

        # Set AMS state correctly for response
        state = struct.unpack("<H", request.ams_header.state_flags)[0]
        state = state | 0x0001  # Set response flag
        state = struct.pack("<H", state)

        # Handle request
        if command_id == constants.ADSCOMMAND_READDEVICEINFO:
            logger.info("Command received: READ_DEVICE_INFO")

            # Create dummy response: version 1.2.3, device name 'TestServer'
            major_version = "\x01".encode("utf-8")
            minor_version = "\x02".encode("utf-8")
            version_build = "\x03\x00".encode("utf-8")
            device_name = "TestServer\x00".encode("utf-8")

            response_content = (
                    major_version + minor_version + version_build + device_name
            )

        elif command_id == constants.ADSCOMMAND_READ:
            logger.info("Command received: READ")
            # Parse requested data length
            response_length = \
                struct.unpack("<I", request.ams_header.data[8:12])[0]
            # Create response of repeated 0x0F with a null terminator for strings
            response_value = (
                    ("\x0F" * (response_length - 1)) + "\x00").encode(
                "utf-8")
            response_content = struct.pack("<I", len(
                response_value)) + response_value

        elif command_id == constants.ADSCOMMAND_WRITE:
            logger.info("Command received: WRITE")
            # No response data required
            response_content = "".encode("utf-8")

        elif command_id == constants.ADSCOMMAND_READSTATE:
            logger.info("Command received: READ_STATE")
            ads_state = struct.pack("<H", constants.ADSSTATE_RUN)
            # I don't know what an appropriate value for device state is.
            # I suspect it may be unused..
            device_state = struct.pack("<H", 0)

            response_content = ads_state + device_state

        elif command_id == constants.ADSCOMMAND_WRITECTRL:
            logger.info("Command received: WRITE_CONTROL")
            # No response data required
            response_content = "".encode("utf-8")

        elif command_id == constants.ADSCOMMAND_ADDDEVICENOTE:
            logger.info("Command received: ADD_DEVICE_NOTIFICATION")
            handle = ("\x0F" * 4).encode("utf-8")
            response_content = handle

        elif command_id == constants.ADSCOMMAND_DELDEVICENOTE:
            logger.info("Command received: DELETE_DEVICE_NOTIFICATION")
            # No response data required
            response_content = "".encode("utf-8")

        elif command_id == constants.ADSCOMMAND_DEVICENOTE:
            logger.info("Command received: DEVICE_NOTIFICATION")
            # No response data required
            response_content = "".encode("utf-8")

        elif command_id == constants.ADSCOMMAND_READWRITE:
            logger.info("Command received: READ_WRITE")
            # parse the request
            index_group = struct.unpack("<I", request.ams_header.data[:4])[0]
            response_length = \
                struct.unpack("<I", request.ams_header.data[8:12])[0]
            write_length = struct.unpack("<I", request.ams_header.data[12:16])[
                0]
            write_data = request.ams_header.data[16: (16 + write_length)]

            if index_group == constants.ADSIGRP_SYM_INFOBYNAMEEX:
                # Pack the structure in the same format as SAdsSymbolEntry.
                # Only 'EntrySize' (first field) and Type will be filled.
                # Use fixed UINT8 type
                if "str_" in write_data.decode():
                    response_value = struct.pack(
                        "<IIIIIIHHH", 30, 0, 0, 5, constants.ADST_STRING, 0, 0,
                        0, 0
                    )
                # Non-existent type
                elif "no_type" in write_data.decode():
                    response_value = struct.pack(
                        "<IIIIIIHHH", 30, 0, 0, 5, 1, 0, 0, 0, 0
                    )
                # Array
                elif "ar_" in write_data.decode():
                    response_value = struct.pack(
                        "<IIIIIIHHH", 30, 0, 0, 2, constants.ADST_UINT8, 0, 0,
                        0, 0
                    )
                else:
                    logger.info("Packing ADST_UINT8...")
                    response_value = struct.pack(
                        "<IIIIIIHHH", 30, 0, 0, 1, constants.ADST_UINT8, 0, 0,
                        0, 0
                    )

            elif index_group == constants.ADSIGRP_SUMUP_READ:
                n_reads = len(write_data) // 12
                fmt = "<" + n_reads * "I"
                vals: List[Union[int, bytes]] = [0 for _ in range(n_reads)]

                for i in range(n_reads):
                    buf = write_data[i * 12 + 8:i * 12 + 12]
                    is_str = struct.unpack("<I", buf)[0] == 5

                    if is_str:
                        fmt += "5s"
                        vals.append(b"test\x00")
                    else:
                        fmt += "B"
                        vals.append(i + 1)
                response_value = struct.pack(fmt, *vals)

            elif index_group == constants.ADSIGRP_SUMUP_WRITE:
                n_writes = len(write_data) // 12
                fmt = "<" + n_writes * "I"
                vals = n_writes * [0]
                response_value = struct.pack(fmt, *vals)

            elif response_length > 0:
                # Create response of repeated 0x0F with a null terminator for strings
                response_value = (
                        ("\x0F" * (response_length - 1)) + "\x00").encode(
                    "utf-8"
                )
            else:
                response_value = b""

            response_content = struct.pack("<I", len(
                response_value)) + response_value

        else:
            logger.info("Unknown Command: {0}".format(hex(command_id)))
            # Set error code to 'unknown command ID'
            error_code = "\x08\x00\x00\x00".encode("utf-8")
            return AmsResponseData(state, error_code, "".encode("utf-8"))

        # Set no error in response
        error_code = ("\x00" * 4).encode("utf-8")
        response_data = error_code + response_content

        return AmsResponseData(state, request.ams_header.error_code,
                               response_data)



"""ADS Connection class.

:author: Stefan Lehmann <stlm@posteo.de>
:maintainer: Filippo Boido <filippo.boido@agileautomation.eu> (Agile Automation Technologies GmbH)
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2018-06-11 18:15:53

"""
from __future__ import annotations
import struct
from ctypes import (
    memmove,
    addressof,
    c_ubyte,
    Array,
    Structure,
    sizeof,
    create_string_buffer,
)
from datetime import datetime
from functools import partial
from typing import Optional, Union, Tuple, Any, Type, Callable, Dict, List, Sequence, cast

# noinspection PyUnresolvedReferences
from .constants import (
    ADSIGRP_SYM_UPLOAD,
    ADSIGRP_SYM_UPLOADINFO2,
    ADSIOFFS_DEVDATA_ADSSTATE,
    PLCTYPE_BOOL,
    PLCTYPE_BYTE,
    PLCTYPE_DATE,
    PLCTYPE_DINT,
    PLCTYPE_DT,
    PLCTYPE_DWORD,
    PLCTYPE_INT,
    PLCTYPE_LREAL,
    PLCTYPE_REAL,
    PLCTYPE_SINT,
    PLCTYPE_STRING,
    PLCTYPE_TIME,
    PLCTYPE_TOD,
    PLCTYPE_UDINT,
    PLCTYPE_UINT,
    PLCTYPE_USINT,
    PLCTYPE_WORD,
    PLC_DEFAULT_STRING_SIZE,
    DATATYPE_MAP,
    ADSIGRP_SUMUP_READ,
    ADSIGRP_SUMUP_WRITE,
    ADSIGRP_SYM_VALBYHND,
    MAX_ADS_SUB_COMMANDS,
    ads_type_to_ctype,
    PLCSimpleDataType,
    PLCDataType,
)
from .filetimes import filetime_to_dt
from .pyads_ex import (
    adsAddRoute,
    adsDelRoute,
    adsPortOpenEx,
    adsPortCloseEx,
    adsGetLocalAddressEx,
    adsSyncReadStateReqEx,
    adsSyncReadDeviceInfoReqEx,
    adsSyncWriteControlReqEx,
    adsSyncWriteReqEx,
    adsSyncReadWriteReqEx2,
    adsSyncReadReqEx2,
    adsGetHandle,
    adsGetNetIdForPLC,
    adsGetSymbolInfo,
    adsSumRead,
    adsSumWrite,
    adsReleaseHandle,
    adsSyncReadByNameEx,
    adsSyncWriteByNameEx,
    adsSyncAddDeviceNotificationReqEx,
    adsSyncDelDeviceNotificationReqEx,
    adsSyncSetTimeoutEx,
    type_is_string,
    type_is_wstring,
    ADSError,
)
from .structs import (
    AmsAddr,
    AdsVersion,
    NotificationAttrib,
    SAdsNotificationHeader,
    SAdsSymbolEntry,
)
from .ads import (
    linux,
    StructureDef,
    dict_from_bytes,
    _list_slice_generator,
    _dict_slice_generator,
    bytes_from_dict,
    size_of_structure,
)
from .symbol import AdsSymbol
from .utils import decode_ads


class RpcObject:
    """Proxy object for calling PLC RPC methods using native attribute syntax."""

    def __init__(
            self,
            connection: "Connection",
            object_name: str,
            method_separator: str = "#",
            method_prefixes: Tuple[str, ...] = ("", "m_"),
            method_return_types: Optional[Dict[str, Type["PLCDataType"]]] = None,
            method_parameters: Optional[
                Dict[str, Sequence[Type["PLCDataType"]]]
            ] = None,
    ) -> None:
        self._connection = connection
        self._object_name = object_name
        self._method_separator = method_separator
        self._method_prefixes = method_prefixes
        self._method_return_types: Dict[str, Type["PLCDataType"]] = (
            method_return_types.copy() if method_return_types else {}
        )
        self._method_parameters: Dict[str, Tuple[Type["PLCDataType"], ...]] = (
            {name: tuple(values) for name, values in method_parameters.items()}
            if method_parameters
            else {}
        )
        # If a method is declared in `method_return_types` but no explicit
        # parameter signature is provided, treat it as a zero-argument RPC.
        for method_name in self._method_return_types:
            self._method_parameters.setdefault(method_name, tuple())
        self._return_spec_cache: Dict[
            str, Tuple[Optional[Type["PLCDataType"]], bool]
        ] = {}

    def _candidate_method_names(self, method_name: str) -> List[str]:
        names: List[str] = []
        for prefix in self._method_prefixes:
            if method_name.startswith(prefix):
                candidate = method_name
            else:
                candidate = f"{prefix}{method_name}"
            names.append(f"{self._object_name}{self._method_separator}{candidate}")
        # Keep order but drop duplicates
        return list(dict.fromkeys(names))

    def _get_symbol_info(self, qualified_name: str) -> Optional[SAdsSymbolEntry]:
        """Return cached symbol info for a fully qualified RPC method name."""
        if not self._connection.is_open or self._connection._port is None:
            return None

        cache = self._connection._symbol_info_cache
        info = cache.get(qualified_name)
        if info is not None:
            return info

        try:
            info = adsGetSymbolInfo(
                self._connection._port, self._connection._adr, qualified_name
            )
        except ADSError:
            return None

        cache[qualified_name] = info
        return info

    def set_return_type(self, method_name: str, plc_type: Type["PLCDataType"]) -> None:
        """Register or override the expected return type for a method."""
        self._method_return_types[method_name] = plc_type
        self._return_spec_cache.pop(
            f"{self._object_name}{self._method_separator}{method_name}", None
        )

    def _manual_return_type(self, method_name: str) -> Optional[Type["PLCDataType"]]:
        """Look up manually configured return type for a method name."""
        candidates = {method_name}
        for prefix in self._method_prefixes:
            if method_name.startswith(prefix):
                candidates.add(method_name[len(prefix):])
            else:
                candidates.add(f"{prefix}{method_name}")
        for candidate in candidates:
            plc_type = self._method_return_types.get(candidate)
            if plc_type is not None:
                return plc_type
        return None

    def _infer_return_spec(
            self, qualified_name: str
    ) -> Tuple[Optional[Type["PLCDataType"]], bool]:
        """Infer the PLC return type for a method, falling back to raw bytes."""
        cached = self._return_spec_cache.get(qualified_name)
        if cached is not None:
            return cached

        plc_type: Optional[Type["PLCDataType"]] = None
        coerce_bytes = False

        _, _, method_name = qualified_name.partition(self._method_separator)
        if method_name:
            plc_type = self._manual_return_type(method_name)

        if plc_type is None:
            info = self._get_symbol_info(qualified_name)
            if info is not None:
                plc_type = ads_type_to_ctype.get(info.dataType)
                if plc_type is None:
                    plc_type = AdsSymbol.get_type_from_str(info.symbol_type)
                if plc_type is None and info.size:
                    plc_type = c_ubyte * info.size
                    coerce_bytes = True

        spec = (plc_type, coerce_bytes)
        self._return_spec_cache[qualified_name] = spec
        return spec

    @staticmethod
    def _normalize_control_args(
            args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> Tuple[Any, Optional[Type["PLCDataType"]], Optional[Type["PLCDataType"]], Optional[str]]:
        write_value = kwargs.pop("write_value", None)
        write_type = kwargs.pop("write_type", None)
        return_type = kwargs.pop("return_type", None)
        method_name_override = kwargs.pop("method_name", None)
        if kwargs:
            unknown = ", ".join(sorted(kwargs.keys()))
            raise TypeError(f"Unknown keyword argument(s): {unknown}")

        if len(args) > 3:
            raise TypeError("Expected up to 3 positional args: value, write_type, return_type")
        if len(args) >= 1:
            if write_value is not None:
                raise TypeError("write_value provided both positionally and as keyword")
            write_value = args[0]
        if len(args) >= 2:
            if write_type is not None:
                raise TypeError("write_type provided both positionally and as keyword")
            write_type = args[1]
        if len(args) == 3:
            if return_type is not None:
                raise TypeError("return_type provided both positionally and as keyword")
            return_type = args[2]

        return write_value, write_type, return_type, method_name_override

    def _parameter_types_for(self, method_name: str) -> Tuple[Type["PLCDataType"], ...]:
        candidates = {method_name}
        for prefix in self._method_prefixes:
            if method_name.startswith(prefix):
                candidates.add(method_name[len(prefix):])
            else:
                candidates.add(f"{prefix}{method_name}")
        for candidate in candidates:
            values = self._method_parameters.get(candidate)
            if values is not None:
                return values
        return tuple()

    @staticmethod
    def _pack_single_argument(value: Any, plc_type: Type["PLCDataType"]) -> bytes:
        if type_is_string(plc_type):
            return value.encode("utf-8") + b"\x00"
        if type_is_wstring(plc_type):
            return value.encode("utf-16-le")

        if type(plc_type).__name__ == "PyCArrayType":
            data = plc_type(*value)
        elif type(value) is plc_type:
            data = value
        else:
            data = plc_type(value)

        raw = (c_ubyte * sizeof(data))()
        memmove(raw, addressof(data), sizeof(data))
        return bytes(raw)

    def _pack_method_args(
            self,
            method_name: str,
            args: Tuple[Any, ...],
    ) -> Tuple[Any, Optional[Type["PLCDataType"]]]:
        parameter_types = self._parameter_types_for(method_name)

        if not parameter_types:
            if len(args) == 0:
                return None, None
            if len(args) == 1:
                raise TypeError(
                    "Parameter types are not configured for this method. "
                    "Pass `method_parameters=` to get_object, "
                    "or call with explicit write_type."
                )
            raise TypeError(
                "Parameter types are not configured for this method. "
                "Use explicit write_type or configure method parameters."
            )

        if len(args) != len(parameter_types):
            raise TypeError(
                f"{method_name} expects {len(parameter_types)} parameter(s), "
                f"got {len(args)}."
            )

        payload = bytearray()
        for value, plc_type in zip(args, parameter_types):
            payload.extend(self._pack_single_argument(value, plc_type))

        if not payload:
            return None, None

        write_type = c_ubyte * len(payload)
        return list(payload), write_type

    def __getattr__(self, method_name: str) -> Callable[..., Any]:
        if method_name.startswith("_"):
            raise AttributeError(method_name)

        def _invoke(*args: Any, **kwargs: Any) -> Any:
            control_keys = {"write_value", "write_type", "return_type", "method_name"}
            has_control_args = any(key in kwargs for key in control_keys)
            if has_control_args:
                write_value, write_type, return_type, method_name_override = self._normalize_control_args(args, kwargs)
            else:
                if kwargs:
                    unknown = ", ".join(sorted(kwargs.keys()))
                    raise TypeError(f"Unknown keyword argument(s): {unknown}")
                is_legacy_positional = (
                    len(args) <= 3
                    and (
                        len(args) == 0
                        or (len(args) >= 2 and isinstance(args[1], type))
                    )
                    and not self._parameter_types_for(method_name)
                )
                if is_legacy_positional:
                    write_value = args[0] if len(args) >= 1 else None
                    write_type = args[1] if len(args) >= 2 else None
                    return_type = args[2] if len(args) == 3 else None
                    method_name_override = None
                else:
                    write_value, write_type = self._pack_method_args(method_name, args)
                    return_type = None
                    method_name_override = None

            if method_name_override is not None:
                return self._connection.call_rpc_method(
                    method_name_override,
                    return_type=return_type,
                    write_value=write_value,
                    write_type=write_type,
                )

            last_error: Optional[ADSError] = None
            for candidate_name in self._candidate_method_names(method_name):
                resolved_return_type = return_type
                coerce_bytes = False
                if resolved_return_type is None:
                    resolved_return_type, coerce_bytes = self._infer_return_spec(
                        candidate_name
                    )
                try:
                    result = self._connection.call_rpc_method(
                        candidate_name,
                        return_type=resolved_return_type,
                        write_value=write_value,
                        write_type=write_type,
                    )
                except ADSError as exc:
                    if exc.err_code != 1808:  # symbol not found
                        raise
                    last_error = exc
                    continue
                if coerce_bytes and isinstance(result, list):
                    return bytes(result)
                return result
            if last_error is not None:
                raise last_error
            raise AttributeError(method_name)

        return _invoke


class Connection(object):
    """Class for managing the connection to an ADS device.

    :ivar str ams_net_id: AMS net id of the remote device
    :ivar int ams_net_port: port of the remote device
    :ivar str ip_address: the ip address of the device

    :note: If no IP address is given the ip address is automatically set
        to first 4 parts of the Ams net id.

    """

    def __init__(
            self, ams_net_id: str = None, ams_net_port: int = None,
            ip_address: str = None
    ) -> None:
        self._port = None  # type: Optional[int]
        self._adr = AmsAddr(ams_net_id, ams_net_port)
        self._open = False
        if ip_address is None:
            if ams_net_id is None:
                raise TypeError("Must provide an IP or net ID")
            self.ip_address = ".".join(ams_net_id.split(".")[:4])
        else:
            self.ip_address = ip_address
        self.ams_net_id = ams_net_id
        self.ams_net_port = ams_net_port
        self._notifications = {}  # type: Dict[int, str]
        self._symbol_info_cache: Dict[str, SAdsSymbolEntry] = {}

    @property
    def ams_netid(self) -> str:
        return self._adr.netid

    @ams_netid.setter
    def ams_netid(self, value: str) -> None:
        if self._open:
            raise AttributeError(
                "Setting netid is not allowed while connection is open."
            )
        self._adr.netid = value

    @property
    def ams_port(self) -> int:
        return self._adr.port

    @ams_port.setter
    def ams_port(self, value: int) -> None:
        if self._open:
            raise AttributeError(
                "Setting port is not allowed while connection is open."
            )
        self._adr.port = value

    def __enter__(self) -> "Connection":
        """Open on entering with-block."""
        self.open()
        return self

    def __exit__(self, _type: Type, _val: Any, _traceback: Any) -> None:
        """Close on leaving with-block."""
        self.close()

    def __del__(self) -> None:
        """Class destructor.

        Make sure to close the connection when an instance runs out of scope.
        """
        # If the connection is already closed, nothing new will happen
        self.close()

    def _query_plc_datatype_from_name(self, data_name: str,
                                      cache_symbol_info: bool) -> Type:
        """Return the plc_datatype by reading SymbolInfo from the target.

        If cache_symbol_info is True then the SymbolInfo will be cached and adsGetSymbolInfo
        will only used once.

        """
        if cache_symbol_info:
            info = self._symbol_info_cache.get(data_name)
            if info is None:
                info = adsGetSymbolInfo(self._port, self._adr, data_name)
                self._symbol_info_cache[data_name] = info
        else:
            info = adsGetSymbolInfo(self._port, self._adr, data_name)
        return AdsSymbol.get_type_from_str(info.symbol_type)

    def open(self) -> None:
        """Connect to the TwinCAT message router."""
        if self._open:
            return

        if self.ams_net_id is None:
            self.ams_net_id = adsGetNetIdForPLC(self.ip_address)
            self._adr = AmsAddr(self.ams_net_id, self.ams_net_port)
        self._port = adsPortOpenEx()

        if linux:
            try:
                adsAddRoute(self._adr.netIdStruct(), self.ip_address)
            except ADSError:
                adsPortCloseEx(self._port)
                self._port = None
                raise

        self._open = True

    def close(self) -> None:
        """:summary: Close the connection to the TwinCAT message router."""
        if not self._open:
            return

        if linux:
            adsDelRoute(self._adr.netIdStruct())

        if self._port is not None:
            adsPortCloseEx(self._port)
            self._port = None

        self._open = False

    def get_local_address(self) -> Optional[AmsAddr]:
        """Return the local AMS-address and the port number.

        :rtype: AmsAddr

        """
        if self._port is not None:
            return adsGetLocalAddressEx(self._port)

        return None

    def read_state(self) -> Optional[Tuple[int, int]]:
        """Read the current ADS-state and the machine-state.

        Read the current ADS-state and the machine-state from the ADS-server.

        :rtype: (int, int)
        :return: adsState, deviceState

        """
        if self._port is not None:
            return adsSyncReadStateReqEx(self._port, self._adr)

        return None

    def write_control(
            self, ads_state: int, device_state: int, data: Any, plc_datatype: Type
    ) -> None:
        """Change the ADS state and the machine-state of the ADS-server.

        :param int ads_state: new ADS-state, according to ADSTATE constants
        :param int device_state: new machine-state
        :param data: additional data
        :param int plc_datatype: datatype, according to PLCTYPE constants

        :note: Despite changing the ADS-state and the machine-state it is
            possible to send additional data to the ADS-server. For current
            ADS-devices additional data is not progressed.
            Every ADS-device is able to communicate its current state to other
            devices. There is a difference between the device-state and the
            state of the ADS-interface (AdsState). The possible states of an
            ADS-interface are defined in the ADS-specification.

        """
        if self._port is not None:
            return adsSyncWriteControlReqEx(
                self._port, self._adr, ads_state, device_state, data, plc_datatype
            )

    def read_device_info(self) -> Optional[Tuple[str, AdsVersion]]:
        """Read the name and the version number of the ADS-server.

        :rtype: string, AdsVersion
        :return: device name, version

        """
        if self._port is not None:
            return adsSyncReadDeviceInfoReqEx(self._port, self._adr)

        return None

    def write(
            self, index_group: int, index_offset: int, value: Any,
            plc_datatype: Type["PLCDataType"]
    ) -> None:
        """Send data synchronous to an ADS-device.

        :param int index_group: PLC storage area, according to the INDEXGROUP
            constants
        :param int index_offset: PLC storage address
        :param Any value: value to write to the storage address of the PLC
        :param Type["PLCDataType"] plc_datatype: type of the data given to the PLC,
            according to PLCTYPE constants

        """
        if self._port is not None:
            return adsSyncWriteReqEx(
                self._port, self._adr, index_group, index_offset, value, plc_datatype
            )

    def read_write(
            self,
            index_group: int,
            index_offset: int,
            plc_read_datatype: Optional[Type["PLCDataType"]],
            value: Any,
            plc_write_datatype: Optional[Type["PLCDataType"]],
            return_ctypes: bool = False,
            check_length: bool = True,
    ) -> Any:
        """Read and write data synchronous from/to an ADS-device.

        :param int index_group: PLC storage area, according to the INDEXGROUP
            constants
        :param int index_offset: PLC storage address
        :param Type["PLCDataType"] plc_read_datatype: type of the data given to the PLC to respond to,
            according to PLCTYPE constants, or None to not read anything
        :param value: value to write to the storage address of the PLC
        :param Type["PLCDataType"] plc_write_datatype: type of the data given to the PLC, according to
            PLCTYPE constants, or None to not write anything
        :param bool return_ctypes: return ctypes instead of python types if True (default: False)
        :param bool check_length: check whether the amount of bytes read matches the size
            of the read data type (default: True)
        :return: value: **value**

        """
        if self._port is not None:
            return adsSyncReadWriteReqEx2(
                self._port,
                self._adr,
                index_group,
                index_offset,
                plc_read_datatype,
                value,
                plc_write_datatype,
                return_ctypes,
                check_length,
            )

        return None

    def read(
            self,
            index_group: int,
            index_offset: int,
            plc_datatype: Type["PLCDataType"],
            return_ctypes: bool = False,
            check_length: bool = True,
    ) -> Any:
        """Read data synchronous from an ADS-device.

        :param int index_group: PLC storage area, according to the INDEXGROUP
            constants
        :param int index_offset: PLC storage address
        :param Type["PLCDataType"] plc_datatype: type of the data given to the PLC, according
            to PLCTYPE constants
        :param bool return_ctypes: return ctypes instead of python types if True
            (default: False)
        :param bool check_length: check whether the amount of bytes read matches the size
            of the read data type (default: True)
        :return: value

        """
        if index_group is None or not isinstance(index_group, int):
            raise TypeError('index_group: integer is required')
        if index_offset is None or not isinstance(index_offset, int):
            raise TypeError('index_offset: integer is required')
        if self._port is not None:
            return adsSyncReadReqEx2(
                self._port,
                self._adr,
                index_group,
                index_offset,
                plc_datatype,
                return_ctypes,
                check_length,
            )

        return None

    def get_symbol(
            self,
            name: Optional[str] = None,
            index_group: Optional[int] = None,
            index_offset: Optional[int] = None,
            plc_datatype: Optional[Union[Type["PLCDataType"], str]] = None,
            comment: Optional[str] = None,
            auto_update: bool = False,
            structure_def: Optional["StructureDef"] = None,
            array_size: Optional[int] = 1,
    ) -> AdsSymbol:
        """Create a symbol instance

        Specify either the variable name or the index_group **and**
        index_offset so the symbol can be located.
        If the name was specified but not all other attributes were,
        the other attributes will be looked up from the connection.
        `data_type` can be a PLCTYPE constant or  a string representing
        a PLC type (e.g. 'LREAL').

        :param str name:
        :param Optional[int] index_group:
        :param Optional[int] index_offset:
        :param plc_datatype: type of the  PLC variable, according
            to PLCTYPE constants
        :param str comment: comment
        :param bool auto_update: Create notification to update buffer (same as
            `set_auto_update(True)`)
        :param Optional["StructureDef"] structure_def: special tuple defining the structure and
            types contained within it according to PLCTYPE constants, must match
            the structure defined in the PLC, PLC structure must be defined with
            {attribute 'pack_mode' :=  '1'}
        :param Optional[int] array_size: size of array if reading array of structure, defaults to 1

        Expected input example for structure_def:

        .. code:: python

            structure_def = (
                ('rVar', pyads.PLCTYPE_LREAL, 1),
                ('sVar', pyads.PLCTYPE_STRING, 2, 35),
                ('SVar1', pyads.PLCTYPE_STRING, 1),
                ('rVar1', pyads.PLCTYPE_REAL, 1),
                ('iVar', pyads.PLCTYPE_DINT, 1),
                ('iVar1', pyads.PLCTYPE_INT, 3),
            )

            # i.e ('Variable Name', variable type, arr size (1 if not array),
            # length of string (if defined in PLC))

        """

        return AdsSymbol(self, name, index_group, index_offset, plc_datatype,
                         comment, auto_update=auto_update, structure_def=structure_def,
                         array_size=array_size)

    def get_all_symbols(self) -> List[AdsSymbol]:
        """Read all symbols from an ADS-device.

        :return: List of AdsSymbols
        """
        symbols = []
        if self._port is not None:
            symbol_size_msg = self.read(
                ADSIGRP_SYM_UPLOADINFO2,
                ADSIOFFS_DEVDATA_ADSSTATE,
                PLCTYPE_STRING,
                return_ctypes=True,
            )
            sym_count = struct.unpack("I", symbol_size_msg[0:4])[0]
            sym_list_length = struct.unpack("I", symbol_size_msg[4:8])[0]

            data_type_creation_fn: Type = cast("Type", partial(create_string_buffer,
                                                               sym_list_length))
            symbol_list_msg = self.read(
                ADSIGRP_SYM_UPLOAD,
                ADSIOFFS_DEVDATA_ADSSTATE,
                data_type_creation_fn,
                return_ctypes=True,
            )

            ptr = 0

            for idx in range(sym_count):
                read_length, index_group, index_offset = struct.unpack(
                    "III", symbol_list_msg[ptr + 0: ptr + 12]
                )
                name_length, type_length, comment_length = struct.unpack(
                    "HHH", symbol_list_msg[ptr + 24: ptr + 30]
                )

                name_start_ptr = ptr + 30
                name_end_ptr = name_start_ptr + name_length
                type_start_ptr = name_end_ptr + 1
                type_end_ptr = type_start_ptr + type_length
                comment_start_ptr = type_end_ptr + 1
                comment_end_ptr = comment_start_ptr + comment_length

                name = decode_ads(symbol_list_msg[name_start_ptr:name_end_ptr])
                symbol_type = decode_ads(symbol_list_msg[type_start_ptr:type_end_ptr])
                comment = decode_ads(symbol_list_msg[comment_start_ptr:comment_end_ptr])

                ptr = ptr + read_length
                symbol = AdsSymbol(plc=self, name=name,
                                   index_group=index_group,
                                   index_offset=index_offset,
                                   symbol_type=symbol_type, comment=comment)
                symbols.append(symbol)
        return symbols

    def get_object(
            self,
            object_name: str,
            method_separator: str = "#",
            method_prefixes: Tuple[str, ...] = ("", "m_"),
            method_return_types: Optional[Dict[str, Type["PLCDataType"]]] = None,
            method_parameters: Optional[
                Dict[str, Sequence[Type["PLCDataType"]]]
            ] = None,
    ) -> RpcObject:
        """Create a PLC object proxy for native RPC method calls.

        Example:
            rpc = plc.get_object(
                "GVL.fbTestRemoteMethodCall",
                method_return_types={"m_iSum": pyads.PLCTYPE_INT},
                method_parameters={"m_iSum": [pyads.PLCTYPE_INT, pyads.PLCTYPE_INT]},
            )
            result = rpc.m_iSum(5, 5)

        Method names are resolved in order using ``method_prefixes``.
        The default supports both ``doCalc`` and ``m_doCalc`` style RPC names.
        """
        return RpcObject(
            connection=self,
            object_name=object_name,
            method_separator=method_separator,
            method_prefixes=method_prefixes,
            method_return_types=method_return_types,
            method_parameters=method_parameters,
        )

    def call_rpc_method(
            self,
            method_name: str,
            return_type: Optional[Type["PLCDataType"]] = None,
            write_value: Any = None,
            write_type: Optional[Type["PLCDataType"]] = None,
    ) -> Any:
        """Call a PLC RPC method by ADS handle and release the handle afterwards."""
        handle = self.get_handle(method_name)
        if handle is None:
            return None

        try:
            return self.read_write(
                ADSIGRP_SYM_VALBYHND,
                handle,
                return_type,
                write_value,
                write_type,
            )
        finally:
            self.release_handle(handle)

    def get_handle(self, data_name: str) -> Optional[int]:
        """Get the handle of the PLC-variable, handles obtained using this
         method should be released using method 'release_handle'.

        :param string data_name: data name

        :rtype: int
        :return: int: PLC-variable handle
        """
        if self._port is not None:
            return adsGetHandle(self._port, self._adr, data_name)

        return None

    def release_handle(self, handle: int) -> None:
        """ Release handle of a PLC-variable.

        :param int handle: handle of PLC-variable to be released
        """
        if self._port is not None:
            adsReleaseHandle(self._port, self._adr, handle)

    def read_by_name(
            self,
            data_name: str,
            plc_datatype: Optional[Type["PLCDataType"]] = None,
            return_ctypes: bool = False,
            handle: Optional[int] = None,
            check_length: bool = True,
            cache_symbol_info: bool = True,
    ) -> Any:
        """Read data synchronous from an ADS-device from data name.

        :param string data_name: data name,  can be empty string if handle is used
        :param Optional[Type["PLCDataType"]] plc_datatype: type of the data given to the PLC, according
            to PLCTYPE constants, if None the datatype will be read from the target
            with adsGetSymbolInfo (default: None)
        :param bool return_ctypes: return ctypes instead of python types if True
            (default: False)
        :param int handle: PLC-variable handle, pass in handle if previously
            obtained to speed up reading (default: None)
        :param bool check_length: check whether the amount of bytes read matches the size
            of the read data type (default: True)
        :param bool cache_symbol_info: when True, symbol info will be cached for
            future reading, only relevant if plc_datatype is None (default: True)
        :return: value: **value**
        """
        if not self._port:
            return

        if plc_datatype is None:
            plc_datatype = self._query_plc_datatype_from_name(data_name,
                                                              cache_symbol_info)

        return adsSyncReadByNameEx(
            self._port,
            self._adr,
            data_name,
            plc_datatype,
            return_ctypes=return_ctypes,
            handle=handle,
            check_length=check_length,
        )

    def read_list_by_name(
            self,
            data_names: List[str],
            cache_symbol_info: bool = True,
            ads_sub_commands: int = MAX_ADS_SUB_COMMANDS,
            structure_defs: Optional[Dict[str, StructureDef]] = None,
    ) -> Dict[str, Any]:
        """Read a list of variables.

        Will split the read into multiple ADS calls in chunks of ads_sub_commands by default.

        MAX_ADS_SUB_COMMANDS comes from Beckhoff recommendation:
        https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_adsdll2/9007199379576075.html&id=9180083787138954512

        :param List[str] data_names: list of variable names to be read
        :param bool cache_symbol_info: when True, symbol info will be cached for future reading
        :param int ads_sub_commands: Max number of ADS-Sub commands used to read the variables in a single ADS call.
            A larger number can be used but may jitter the PLC execution!
        :param Optional[Dict[str, StructureDef]] structure_defs: for structured variables, optional mapping of
            data name to special tuple defining the structure and types contained within it according to PLCTYPE constants
        :return adsSumRead: A dictionary containing variable names from data_names as keys and values read from PLC for each variable
        :rtype: Dict[str, Any]

        """
        if structure_defs is None:
            structure_defs = {}

        if cache_symbol_info:
            new_items = [i for i in data_names if i not in self._symbol_info_cache]
            new_cache = {
                i: adsGetSymbolInfo(self._port, self._adr, i) for i in new_items
            }
            self._symbol_info_cache.update(new_cache)
            data_symbols = {i: self._symbol_info_cache[i] for i in data_names}
        else:
            data_symbols = {
                i: adsGetSymbolInfo(self._port, self._adr, i) for i in data_names
            }

        def sum_read(port: int, adr: AmsAddr, data_names: List[str],
                     data_symbols: Dict) -> Dict[str, str]:
            result = adsSumRead(port, adr, data_names, data_symbols,
                                list(structure_defs.keys()))  # type: ignore

            for data_name, structure_def in structure_defs.items():  # type: ignore
                if data_name in result:
                    result[data_name] = dict_from_bytes(result[data_name],
                        structure_def)

            return result

        if len(data_names) <= ads_sub_commands:
            return sum_read(self._port, self._adr, data_names, data_symbols)

        return_data: Dict[str, Any] = {}
        for data_names_slice in _list_slice_generator(data_names, ads_sub_commands):
            return_data.update(
                sum_read(self._port, self._adr, data_names_slice, data_symbols)
            )
        return return_data

    def read_structure_by_name(
            self,
            data_name: str,
            structure_def: StructureDef,
            array_size: Optional[int] = 1,
            structure_size: Optional[int] = None,
            handle: Optional[int] = None,
    ) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:
        """Read a structure of multiple types.

        :param string data_name: data name
        :param tuple structure_def: special tuple defining the structure and
            types contained within it according to PLCTYPE constants, must match
            the structure defined in the PLC, PLC structure must be defined with
            {attribute 'pack_mode' :=  '1'}
        :param Optional[int] array_size: size of array if reading array of structure, defaults to 1
        :param Optional[int] structure_size: size of structure if known by previous use of
            size_of_structure, defaults to None
        :param Optional[int] handle: PLC-variable handle, pass in handle if previously
            obtained to speed up reading, defaults to None
        :return: values_dict: ordered dictionary of all values corresponding to the structure
            definition

        Expected input example for structure_def:

        .. code:: python

            structure_def = (
                ('rVar', pyads.PLCTYPE_LREAL, 1),
                ('sVar', pyads.PLCTYPE_STRING, 2, 35),
                ('SVar1', pyads.PLCTYPE_STRING, 1),
                ('rVar1', pyads.PLCTYPE_REAL, 1),
                ('iVar', pyads.PLCTYPE_DINT, 1),
                ('iVar1', pyads.PLCTYPE_INT, 3),
            )

            # i.e ('Variable Name', variable type, arr size (1 if not array),
            # length of string (if defined in PLC))

        """
        if structure_size is None:
            structure_size = size_of_structure(structure_def * array_size)
        values = self.read_by_name(data_name, c_ubyte * structure_size, handle=handle)
        if values is not None:
            return dict_from_bytes(values, structure_def, array_size=array_size)

        return None

    def write_by_name(
            self,
            data_name: str,
            value: Any,
            plc_datatype: Optional[Type["PLCDataType"]] = None,
            handle: Optional[int] = None,
            cache_symbol_info: bool = True,
    ) -> None:
        """Send data synchronous to an ADS-device from data name.

        :param string data_name: data name, can be empty string if handle is used
        :param value: value to write to the storage address of the PLC
        :param int plc_datatype: type of the data given to the PLC, according
            to PLCTYPE constants, if None the datatype will be read from the target
            with adsGetSymbolInfo (default: None)
        :param int handle: PLC-variable handle, pass in handle if previously
            obtained to speed up writing (default: None)
        :param bool cache_symbol_info: when True, symbol info will be cached for
            future reading, only relevant if plc_datatype is None (default: True)
        """
        if not self._port:
            return

        if plc_datatype is None:
            plc_datatype = self._query_plc_datatype_from_name(data_name,
                                                              cache_symbol_info)

        return adsSyncWriteByNameEx(
            self._port, self._adr, data_name, value, plc_datatype, handle=handle
        )

    def write_list_by_name(
            self,
            data_names_and_values: Dict[str, Any],
            cache_symbol_info: bool = True,
            ads_sub_commands: int = MAX_ADS_SUB_COMMANDS,
            structure_defs: Optional[Dict[str, StructureDef]] = None,
    ) -> Dict[str, str]:
        """Write a list of variables.

        Will split the write into multiple ADS calls in chunks of ads_sub_commands by default.

        MAX_ADS_SUB_COMMANDS comes from Beckhoff recommendation:
        https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_adsdll2/9007199379576075.html&id=9180083787138954512

        :param data_names_and_values: dictionary of variable names and their values to be written
        :type data_names_and_values: dict[str, Any]
        :param bool cache_symbol_info: when True, symbol info will be cached for future reading
        :param int ads_sub_commands: Max number of ADS-Sub commands used to write the variables in a single ADS call.
            A larger number can be used but may jitter the PLC execution!
        :param dict structure_defs: for structured variables, optional mapping of
            data name to special tuple defining the structure and
            types contained within it according to PLCTYPE constants
        :return adsSumWrite: A dictionary containing variable names from data_names as keys and values return codes for
            each write operation from the PLC
        :rtype: dict(str, str)

        """
        if cache_symbol_info:
            new_items = [
                i
                for i in data_names_and_values.keys()
                if i not in self._symbol_info_cache
            ]
            new_cache = {
                i: adsGetSymbolInfo(self._port, self._adr, i) for i in new_items
            }
            self._symbol_info_cache.update(new_cache)
            data_symbols = {
                i: self._symbol_info_cache[i] for i in data_names_and_values
            }
        else:
            data_symbols = {
                i: adsGetSymbolInfo(self._port, self._adr, i)
                for i in data_names_and_values.keys()
            }

        if structure_defs is None:
            structure_defs = {}
        else:
            data_names_and_values = data_names_and_values.copy()  # copy so the original does not get modified

        for name, structure_def in structure_defs.items():
            data_names_and_values[name] = bytes_from_dict(data_names_and_values[name],
                                                          structure_def)

        structured_data_names = list(structure_defs.keys())

        if len(data_names_and_values) <= ads_sub_commands:
            return adsSumWrite(
                self._port, self._adr, data_names_and_values, data_symbols,
                structured_data_names
            )

        return_data: Dict[str, str] = {}
        for data_names_slice in _dict_slice_generator(data_names_and_values,
                                                      ads_sub_commands):
            return_data.update(
                adsSumWrite(self._port, self._adr, data_names_slice, data_symbols,
                            structured_data_names)
            )
        return return_data

    def write_structure_by_name(
            self,
            data_name: str,
            value: Union[Dict[str, Any], List[Dict[str, Any]]],
            structure_def: StructureDef,
            array_size: Optional[int] = 1,
            structure_size: Optional[int] = None,
            handle: Optional[int] = None,
    ) -> None:
        """Write a structure of multiple types.

        :param str data_name: data name
        :param Union[Dict[str, Any], List[Dict[str, Any]]] value: value to write to the storage address of the PLC
        :param StructureDef structure_def: special tuple defining the structure and
            types contained within it according to PLCTYPE constants, must match
            the structure defined in the PLC, PLC structure must be defined with
            {attribute 'pack_mode' :=  '1'}
        :param Optional[int] array_size: size of array if writing array of structure, defaults to 1
        :param Optional[int] structure_size: size of structure if known by previous use of
            size_of_structure, defaults to None
        :param Optional[int] handle: PLC-variable handle, pass in handle if previously
            obtained to speed up reading, defaults to None

        Expected input example for structure_def:

        .. code:: python

            structure_def = (
                ('rVar', pyads.PLCTYPE_LREAL, 1),
                ('sVar', pyads.PLCTYPE_STRING, 2, 35),
                ('sVar', pyads.PLCTYPE_STRING, 1),
                ('rVar1', pyads.PLCTYPE_REAL, 1),
                ('iVar', pyads.PLCTYPE_DINT, 1),
            )

            # i.e ('Variable Name', variable type, arr size (1 if not array),
            # length of string (if defined in PLC))

        """
        byte_values = bytes_from_dict(value, structure_def)
        if structure_size is None:
            structure_size = size_of_structure(structure_def * array_size)
        return self.write_by_name(
            data_name, byte_values, c_ubyte * structure_size, handle=handle
        )

    def add_device_notification(
            self,
            data: Union[str, Tuple[int, int]],
            attr: NotificationAttrib,
            callback: Callable,
            user_handle: Optional[int] = None,
    ) -> Optional[Tuple[int, int]]:
        """Add a device notification.

        :param Union[str, Tuple[int, int] data: PLC storage address as string or Tuple with index group and offset
        :param pyads.structs.NotificationAttrib attr: object that contains
            all the attributes for the definition of a notification
        :param callback: callback function that gets executed in the event of a notification
        :param user_handle: optional user handle

        :rtype: (int, int)
        :returns: notification handle, user handle

        Save the notification handle and the user handle on creating a
        notification if you want to be able to remove the notification
        later in your code.

        **Usage**:

            >>> import pyads
            >>> from ctypes import sizeof
            >>>
            >>> # Connect to the local TwinCAT PLC
            >>> plc = pyads.Connection('127.0.0.1.1.1', 851)
            >>>
            >>> # Create callback function that prints the value
            >>> def mycallback(notification, data):
            >>>     contents = notification.contents
            >>>     value = next(
            >>>         map(int,
            >>>             bytearray(contents.data)[0:contents.cbSampleSize])
            >>>     )
            >>>     print(value)
            >>>
            >>> with plc:
            >>>     # Add notification with default settings
            >>>     atr = pyads.NotificationAttrib(sizeof(pyads.PLCTYPE_INT))
            >>>     handles = plc.add_device_notification("GVL.myvalue", atr, mycallback)
            >>>
            >>>     # Remove notification
            >>>     plc.del_device_notification(handles)

        Note: the `user_handle` (passed or returned) is the same as the handle returned from
        :meth:`Connection.get_handle()`.

        """
        if self._port is not None:
            notification_handle, user_handle = adsSyncAddDeviceNotificationReqEx(
                self._port, self._adr, data, attr, callback, user_handle
            )
            return notification_handle, user_handle

        return None

    def del_device_notification(
            self, notification_handle: int, user_handle: int
    ) -> None:
        """Remove a device notification.

        :param notification_handle: address of the variable that contains
            the handle of the notification
        :param user_handle: user handle

        """
        if self._port is not None:
            adsSyncDelDeviceNotificationReqEx(
                self._port, self._adr, notification_handle, user_handle
            )

    @property
    def is_open(self) -> bool:
        """Show the current connection state.

        :return: True if connection is open

        """
        return self._open

    def set_timeout(self, ms: int) -> None:
        """Set Timeout."""
        if self._port is not None:
            adsSyncSetTimeoutEx(self._port, ms)

    def notification(
            self, plc_datatype: Optional[Type] = None,
            timestamp_as_filetime: bool = False
    ) -> Callable:
        """Decorate a callback function.

        **Decorator**.

        A decorator that can be used for callback functions in order to
        convert the data of the NotificationHeader into the fitting
        Python type.

        :param plc_datatype: The PLC datatype that needs to be converted. This can
            be any basic PLC datatype or a `ctypes.Structure`.
        :param timestamp_as_filetime: Whether the notification timestamp should be returned
            as `datetime.datetime` (False) or Windows `FILETIME` as originally transmitted
            via ADS (True). Be aware that the precision of `datetime.datetime` is limited to
            microseconds, while FILETIME allows for 100 ns. This may be relevant when using
            task cycle times such as 62.5 µs. Default: False.

        The callback functions need to be of the following type:

        >>> def callback(handle, name, timestamp, value)

        * `handle`: the notification handle
        * `name`: the variable name
        * `timestamp`: the timestamp as datetime value
        * `value`: the converted value of the variable

        **Usage**:

            >>> import pyads
            >>>
            >>> plc = pyads.Connection('172.18.3.25.1.1', 851)
            >>>
            >>>
            >>> @plc.notification(pyads.PLCTYPE_STRING)
            >>> def callback(handle, name, timestamp, value):
            >>>     print(handle, name, timestamp, value)
            >>>
            >>>
            >>> with plc:
            >>>    attr = pyads.NotificationAttrib(20,
            >>>                                    pyads.ADSTRANS_SERVERCYCLE)
            >>>    handles = plc.add_device_notification('GVL.test', attr,
            >>>                                          callback)
            >>>    while True:
            >>>        pass

        """

        def notification_decorator(
                func: Callable[[int, str, Union[datetime, int], Any], None]
        ) -> Callable[[Any, str], None]:
            def func_wrapper(notification: Any, data_name: str) -> None:
                h_notification, timestamp, value = self.parse_notification(
                    notification, plc_datatype, timestamp_as_filetime
                )
                return func(h_notification, data_name, timestamp, value)

            return func_wrapper

        return notification_decorator

    # noinspection PyMethodMayBeStatic
    def parse_notification(
            self,
            notification: Any,
            plc_datatype: Optional[Type],
            timestamp_as_filetime: bool = False,
    ) -> Tuple[int, Union[datetime, int], Any]:
        # noinspection PyTypeChecker
        """Parse a notification.

                        Convert the data of the NotificationHeader into the fitting Python type.

                        :param notification: The notification we recieve from PLC datatype to be
                            converted. This can be any basic PLC datatype or a `ctypes.Structure`.
                        :param plc_datatype: The PLC datatype that needs to be converted. This can
                            be any basic PLC datatype or a `ctypes.Structure`.
                        :param timestamp_as_filetime: Whether the notification timestamp should be returned
                            as `datetime.datetime` (False) or Windows `FILETIME` as originally transmitted
                            via ADS (True). Be aware that the precision of `datetime.datetime` is limited to
                            microseconds, while FILETIME allows for 100 ns. This may be relevant when using
                            task cycle times such as 62.5 µs. Default: False.

                        :rtype: (int, int, Any)
                        :returns: notification handle, timestamp, value

                        **Usage**:

                        >>> import pyads
                        >>> from ctypes import sizeof
                        >>>
                        >>> # Connect to the local TwinCAT PLC
                        >>> plc = pyads.Connection('127.0.0.1.1.1', 851)
                        >>> tag = {"GVL.myvalue": pyads.PLCTYPE_INT}
                        >>>
                        >>> # Create callback function that prints the value
                        >>> def mycallback(notification: SAdsNotificationHeader, data: str) -> None:
                        >>>     data_type = tag[data]
                        >>>     handle, timestamp, value = plc.parse_notification(notification, data_type)
                        >>>     print(value)
                        >>>
                        >>> with plc:
                        >>>     # Add notification with default settings
                        >>>     attr = pyads.NotificationAttrib(sizeof(pyads.PLCTYPE_INT))
                        >>>
                        >>>     handles = plc.add_device_notification("GVL.myvalue", attr, mycallback)
                        >>>
                        >>>     # Remove notification
                        >>>     plc.del_device_notification(handles)
                        """
        contents = notification.contents
        data_size = contents.cbSampleSize
        # Get dynamically sized data array
        data = (c_ubyte * data_size).from_address(
            addressof(contents) + SAdsNotificationHeader.data.offset
        )
        value: Any
        if plc_datatype == PLCTYPE_STRING:
            # read only until null-termination character
            value = bytearray(data).split(b"\0", 1)[0].decode("utf-8")

        elif plc_datatype is not None and issubclass(plc_datatype, Structure):
            value = plc_datatype()
            fit_size = min(data_size, sizeof(value))
            memmove(addressof(value), addressof(data), fit_size)

        elif plc_datatype is not None and issubclass(plc_datatype, Array):
            if data_size == sizeof(plc_datatype):
                value = list(plc_datatype.from_buffer_copy(bytes(data)))
            else:
                # invalid size
                value = None

        elif plc_datatype not in DATATYPE_MAP:
            value = bytearray(data)

        else:
            value = struct.unpack(DATATYPE_MAP[plc_datatype], bytearray(data))[0]

        if timestamp_as_filetime:
            timestamp = contents.nTimeStamp
        else:
            timestamp = filetime_to_dt(contents.nTimeStamp)

        return contents.hNotification, timestamp, value



"""Structs for the work with ADS API.

:author: Stefan Lehmann <stlm@posteo.de>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2018-06-11 18:15:53

"""

import ctypes
from ctypes import Structure, c_ubyte, c_uint16, c_uint32, c_uint64
from typing import Union

from .constants import ADSTRANS_SERVERONCHA


class SAdsVersion(Structure):
    """Struct containing ADS version information."""

    _fields_ = [("version", c_ubyte), ("revision", c_ubyte), ("build", c_uint16)]


class AdsVersion:
    """Contains version number, revision number, build number of the ADS-DLL.

    :ivar int version: version number
    :ivar int revision: revision number
    :ivar int build: build number

    """

    def __init__(self, stAdsVersion: SAdsVersion) -> None:
        """Create new AdsVersion object.

        :param pyads.constants.SAdsVersion stAdsVersion: ctypes structure
            with the version info

        """
        self.version = stAdsVersion.version
        self.revision = stAdsVersion.revision
        self.build = stAdsVersion.build


class SAmsNetId(Structure):
    """Struct with array of 6 bytes used to describe a net id."""

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [("b", c_ubyte * 6)]


class SAmsAddr(Structure):
    """Struct containing the netId and port of an ADS device."""

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [("netId", SAmsNetId), ("port", c_uint16)]


class AmsAddr(object):
    """Wrapper for SAmsAddr-structure to address an ADS device.

    :type _ams_addr: SAmsAddr
    :ivar _ams_addr: ctypes-structure SAmsAddr

    """

    def __init__(self, netid: str = None, port: int = None) -> None:
        """Create a new AmsAddr object by a given netid and port.

        :param netid: NetId of an ADS device
        :param port: port of an ADS device

        """
        self._ams_addr = SAmsAddr()

        if netid is not None:
            self.netid = netid
        if port is not None:
            self.port = port

    def toString(self) -> str:
        """Textual representation of the AMS address.

        :rtype: string
        :return:  textual representation of the AMS address
        """
        return self.netid + ": " + str(self._ams_addr.port)

    # property netid
    @property
    def netid(self) -> str:
        """Netid of the AmsAddress.

        The Netid is always returned as a String. If the NetId is set
        it can be passed as a String or as a SAmsNetId struct.

        """
        return ".".join(map(str, self._ams_addr.netId.b))

    @netid.setter
    def netid(self, value: Union[str, SAmsNetId]) -> None:
        # Check if the value is already an instance of the SAmsNetId struct
        if isinstance(value, SAmsNetId):
            self._ams_addr.netId = value

        # Otherwise, attempt to parse the id as a string
        else:
            id_numbers = list(map(int, value.split(".")))

            if len(id_numbers) != 6:
                raise ValueError("no valid netid")

            # Fill the netId struct with data
            self._ams_addr.netId.b = (c_ubyte * 6)(*id_numbers)

    # property port
    @property
    def port(self) -> int:
        """Port of the AmsAddress object."""
        return self._ams_addr.port

    @port.setter
    def port(self, value: int) -> None:
        self._ams_addr.port = c_uint16(value)

    def amsAddrStruct(self) -> SAmsAddr:
        """Return the c-types structure SAmsAddr."""
        return self._ams_addr

    def netIdStruct(self) -> SAmsNetId:
        """Return the c-types structure SAmsNetId."""
        return self._ams_addr.netId

    def setAdr(self, adrString: str) -> None:
        """Set the AMS-address according to the given IP-address.

        :type adrString: string
        :param adrString: ip-address of an ADS device

        """
        self.netid = adrString

    def __repr__(self) -> str:
        """Return object name."""
        return "<AmsAddress {}:{}>".format(self.netid, self.port)


class NotificationAttrib(object):
    """Notification Attribute."""

    def __init__(
        self, length: int, trans_mode: int = ADSTRANS_SERVERONCHA, max_delay: float = 1e-4, cycle_time: float = 1e-4
    ) -> None:
        """Create a new NotificationAttrib object.

        :param int length: length of the data
        :param int trans_mode: transmission mode
        :param float max_delay: maximum delay in ms
        :param float cycle_time: cycle time in ms

        """
        self._attrib = SAdsNotificationAttrib()
        if length:
            self._attrib.cbLength = length
        if trans_mode:
            self._attrib.nTransMode = trans_mode
        if max_delay:
            self._attrib.nMaxDelay = int(max_delay * 1e4)
        if cycle_time:
            self._attrib.nCycleTime = int(cycle_time * 1e4)

    def notificationAttribStruct(self) -> "SAdsNotificationAttrib":
        """Return the raw struct."""
        return self._attrib

    @property
    def length(self) -> int:
        """Notification data length."""
        return self._attrib.cbLength

    @length.setter
    def length(self, val: int) -> None:
        self._attrib.cbLength = val

    @property
    def trans_mode(self) -> int:
        """Mode of transmission.

        This can be one of the following:

        * ADSTRANS_NOTRANS
        * ADSTRANS_CLIENTCYCLE
        * ADSTRANS_CLIENT1REQ
        * ADSTRANS_SERVERCYCLE
        * ADSTRANS_SERVERONCHA

        """
        return self._attrib.nTransMode

    @trans_mode.setter
    def trans_mode(self, val: int) -> None:
        self._attrib.nTransMode = val

    @property
    def max_delay(self) -> None:
        """Maximum allowed delay between notifications in ms."""
        return self._attrib.nMaxDelay

    @max_delay.setter
    def max_delay(self, val: int) -> None:
        self._attrib.nMaxDelay = val

    @property
    def cycle_time(self) -> int:
        """Notification cycle time in ms for cycle transmission mode."""
        return self._attrib.nCycleTime

    @cycle_time.setter
    def cycle_time(self, val: int) -> None:
        self._attrib.nCycleTime = val
        self._attrib.dwChangeFilter = val

    def __repr__(self) -> str:
        """Return object name."""
        return "<NotificationAttrib {} {} {} {}>".format(
            self.length, self.trans_mode, self.max_delay, self.cycle_time
        )


class _AttribUnion(ctypes.Union):
    _fields_ = [("nCycleTime", c_uint32), ("dwChangeFilter", c_uint32)]


class SAdsNotificationAttrib(Structure):
    """C structure representation of AdsNotificationAttrib."""

    _pack_ = 1
    _layout_ = "ms"
    _anonymous_ = ("AttribUnion",)
    _fields_ = [
        ("cbLength", c_uint32),
        ("nTransMode", c_uint32),
        ("nMaxDelay", c_uint32),
        ("AttribUnion", _AttribUnion),
    ]


# noinspection PyUnresolvedReferences
class SAdsNotificationHeader(Structure):
    """C structure representation of AdsNotificationHeader.

    :ivar hNotification: notification handle
    :ivar nTimeStamp: time stamp in FILETIME format
    :ivar cbSampleSize: number of data bytes
    :ivar data: variable-length data field, get via ctypes.addressof + offset

    """

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [
        ("hNotification", c_uint32),
        ("nTimeStamp", c_uint64),
        ("cbSampleSize", c_uint32),
        ("data", c_ubyte),
    ]


class SAdsSymbolUploadInfo(Structure):
    """C structure representation of AdsSymbolUploadInfo."""

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [("nSymbols", c_uint32), ("nSymSize", c_uint32)]


# noinspection PyUnresolvedReferences
class SAdsSymbolEntry(Structure):
    """ADS symbol information.

    :ivar entryLength: length of complete symbol entry
    :ivar iGroup: indexGroup of symbol: input, output etc.
    :ivar iOffs: indexOffset of symbol
    :ivar size: size of symbol (in bytes, 0=bit)
    :ivar dataType: adsDataType of symbol
    :ivar flags: symbol flags
    :ivar nameLength: length of symbol name
    :ivar typeLength: length of type name
    :ivar commentLength: length of comment

    A complete example could be:

    .. code:: python

        value: 57172            # Current value
        info.entryLength: 88    # Total storage space for this symbol
        info.iGroup: 16448      # Group index
        info.iOffs: 385000      # Offset index inside group
        info.size: 2            # Number of bytes needed for the value
        info.dataType: 18       # Symbol type, in this case
                                  constants.ADST_UINT16 (18)
        info.flags: 8           # TwinCAT byte flags
        info.nameLength: 11     # Number of characters in the name
        info.typeLength: 4      # Number of characters in the PLC string
                                  representation of the type
        info.commentLength: 20  # Number of characters in the comment
        info.stringBuffer: <pyads.structs.c_ubyte_Array_768 object>
                                # Concatenation of all string info
        bytes(info.stringBuffer): b'GVL.counter\x00UINT\x00 Counter (in '
                                  'pulses)\x00\x95\x19\x07\x18\x00\x00\x00\x00'
        bytes(info.stringBuffer).encode(): "GVL.counter UINT Counter (in
                                            pulses)"

        info.name: "GVL.counter"    # The name section from the buffer
        info.symbol_type: "UINT"    # The symbol_type section from the
                                      buffer
        info.comment: " Counter (in pulses)"  # The comment (if any)
    """

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [
        ("entryLength", c_uint32),
        ("iGroup", c_uint32),
        ("iOffs", c_uint32),
        ("size", c_uint32),
        ("dataType", c_uint32),
        ("flags", c_uint32),
        ("nameLength", c_uint16),
        ("typeLength", c_uint16),
        ("commentLength", c_uint16),
        ("stringBuffer", c_ubyte * (256 * 3)),
        # 3 strings contained, with max length 256 each
    ]

    def _get_string(self, offset: int, length: int) -> str:
        """Get portion of the bigger string buffer"""
        return bytes(self.stringBuffer[offset:(offset + length)]) \
            .decode("windows-1252")  # TwinCAT uses Windows encoding for non-ascii characters

    @property
    def name(self) -> str:
        """The symbol name."""
        return self._get_string(0, self.nameLength)

    @property
    def symbol_type(self) -> str:
        """The qualified type name, including the namespace."""
        return self._get_string(self.nameLength + 1, self.typeLength)

    @property
    def comment(self) -> str:
        """User-defined comment."""
        return self._get_string(
            self.nameLength + self.typeLength + 2, self.commentLength
        )


class SAdsSumRequest(Structure):
    """ADS sum request structure.

    :ivar iGroup: indexGroup of request
    :ivar iOffs: indexOffset of request
    :ivar size: size of request
    """

    _pack_ = 1
    _layout_ = "ms"
    _fields_ = [("iGroup", c_uint32), ("iOffset", c_uint32), ("size", c_uint32)]



"""Advanced handler module for testserver.

:author: Stefan Lehmann <stlm@posteo.de>
:license: MIT, see license file or https://opensource.org/licenses/MIT
:created on: 2017-09-15

"""
from typing import Optional, Union, Dict, Tuple, List
import struct
import ctypes
from datetime import datetime

from .handler import AbstractHandler, AmsPacket, AmsResponseData, logger
from pyads import constants, structs
from pyads.filetimes import dt_to_filetime
from pyads.pyads_ex import callback_store


class PLCVariable:
    """Storage item for named data.

    Also include variable type so it can be retrieved later.
    This basically mirrors SAdsSymbolEntry or AdsSymbol, however we want to
    avoid using those directly since they are test subjects.
    """

    handle_count = 10000  # Keep track of the latest awarded handle
    notification_count = 10  # Keep track the latest notification handle

    INDEX_GROUP = 12345
    INDEX_OFFSET_BASE = 10000

    def __init__(
        self,
        name: str,
        value: Union[int, float, bytes],
        ads_type: int,
        symbol_type: str,
        index_group: Optional[int] = None,
        index_offset: Optional[int] = None,
    ) -> None:
        """
        Handle and indices are set by default (to random but safe values)

        :param str name: variable name
        :param bytes value: variable value as bytes
        :param int ads_type: constants.ADST_*
        :param str symbol_type: PLC-style name of type
        :param Optional[int] index_group: set index_group manually
        :param Optional[int] index_offset: set index_offset manually
        """
        self.name = name.strip("\x00")

        # value is stored in binary!
        if isinstance(value, bytes):
            self.value = value
        else:
            # try to pack value according to ads_type
            fmt = constants.DATATYPE_MAP[constants.ads_type_to_ctype[ads_type]]
            self.value = struct.pack(fmt, value)

        self.ads_type = ads_type
        self.symbol_type = symbol_type

        self.handle = PLCVariable.handle_count
        PLCVariable.handle_count += 1

        if index_group is None:
            self.index_group = (
                PLCVariable.INDEX_GROUP
            )  # default value - shouldn't matter much
        else:
            self.index_group = index_group

        if index_offset is None:
            # cheat by using the handle as offset (since we know it will be unique)
            self.index_offset = PLCVariable.INDEX_OFFSET_BASE + self.handle
        else:
            self.index_offset = index_offset

        self.comment: str = ""

        self.notifications: List[int] = []  # List of associated notification handles

    @property
    def size(self) -> int:
        """Return size of value."""
        return len(self.value)

    def get_packed_info(self) -> bytes:
        """Get bytes array of symbol info"""
        if self.comment is None:
            self.comment = ""
        name_bytes = self.name.encode("utf-8")
        symbol_type_bytes = self.symbol_type.encode("utf-8")
        comment_bytes = self.comment.encode("utf-8")

        entry_length = (
            6 * 4
            + 3 * 2
            + len(name_bytes)
            + 1
            + len(symbol_type_bytes)
            + 1
            + len(comment_bytes)
        )

        read_data = (
            struct.pack(
                "<IIIIIIHHH",
                entry_length,  # Number of packed bytes
                self.index_group,
                self.index_offset,
                self.size,
                self.ads_type,
                0,  # Flags
                len(name_bytes),
                len(symbol_type_bytes),
                len(comment_bytes),
            )
            + name_bytes
            + b"\x20"
            + symbol_type_bytes
            + b"\x20"
            + comment_bytes
        )

        return read_data

    def write(self, value: bytes, request: AmsPacket = None):
        """Update the variable value, respecting notifications"""

        if self.value != value:
            if self.notifications:

                header = structs.SAdsNotificationHeader()
                header.hNotification = 0
                # Include the detected local timezone to the datetime-object:
                header.nTimeStamp = dt_to_filetime(datetime.now().astimezone())
                header.cbSampleSize = len(value)

                # Perform byte-write into the header
                dst = ctypes.addressof(header) + structs.SAdsNotificationHeader.data.offset
                ctypes.memmove(dst, value, len(value))

                for notification_handle in self.notifications:

                    # It's hard to guess the exact AmsAddr from here, so instead
                    # ignore the address and search for the note_handle

                    for key, func in callback_store.items():

                        # callback_store is keyed by (AmsAddr, int)
                        if key[1] != notification_handle:
                            continue

                        header.hNotification = notification_handle
                        addr = key[0]

                        # Call c-wrapper for user callback
                        func(addr.amsAddrStruct(), header, 0)

        self.value = value

    def register_notification(self) -> int:
        """Register a new notification."""

        handle = self.notification_count
        self.notifications.append(handle)
        self.notification_count += 1
        return handle

    def unregister_notification(self, handle: int = None):
        """Unregister a notification.

        :param handle: Set to `None` (default) to unregister all notifications
        """

        if handle is None:
            self.notifications = []
        else:
            if handle in self.notifications:
                self.notifications.remove(handle)


class AdvancedHandler(AbstractHandler):
    """The advanced handler allows to store and restore data.

    The advanced handler allows to store and restore data via read, write and
    read_write functions. There is a storage area for each symbol. The
    purpose of this handler to test read/write access and test basic
    interaction.
    Variables can be read/write through indices, name and handle.

    An error will be thrown when an attempt is made to read from a
    non-existent variable. You can either: i) write the variable first (it
    is implicitly created) or ii) create the variable yourself and place it
    in the handler.
    Note that the variable type cannot be set correctly in the implicit
    creation! (It will default to UINT16.) Use explicit creation if a
    non-default type is important.
    """

    def __init__(self) -> None:
        self._data: Dict[Tuple[int, int], PLCVariable] = {}
        # This will be our variables database
        # We won't both with indexing it by handle or name, speed is not
        # important. We store by group + offset index and will have to
        # search inefficiently for name or handle. (Unlike real ADS!)

        self.reset()

    def reset(self) -> None:
        """Clear saved variables in handler"""
        self._data = {}

    def handle_request(self, request: AmsPacket) -> AmsResponseData:
        """Handle incoming requests and create a response."""
        # Extract command id from the request
        command_id_bytes = request.ams_header.command_id
        command_id = struct.unpack("<H", command_id_bytes)[0]

        # Set AMS state correctly for response
        state = struct.unpack("<H", request.ams_header.state_flags)[0]
        state = state | 0x0001  # Set response flag
        state = struct.pack("<H", state)

        def handle_read_device_info() -> bytes:
            """Create dummy response: version 1.2.3, device name 'TestServer'."""
            logger.info("Command received: READ_DEVICE_INFO")

            major_version = "\x01".encode("utf-8")
            minor_version = "\x02".encode("utf-8")
            version_build = "\x03\x00".encode("utf-8")
            device_name = "TestServer\x00".encode("utf-8")

            response_content = (
                major_version + minor_version + version_build + device_name
            )

            return response_content

        def handle_read() -> bytes:
            """Handle read request."""
            data = request.ams_header.data

            index_group = struct.unpack("<I", data[:4])[0]
            index_offset = struct.unpack("<I", data[4:8])[0]
            plc_datatype = struct.unpack("<I", data[8:12])[0]

            logger.info(
                (
                    "Command received: READ (index group={}, index offset={}, "
                    "data length={})"
                ).format(hex(index_group), hex(index_offset), plc_datatype)
            )

            # value by handle is demanded return from named data store
            if index_group == constants.ADSIGRP_SYM_VALBYHND:
                response_value = self.get_variable_by_handle(index_offset).value

            elif index_group == constants.ADSIGRP_SYM_UPLOADINFO2:
                symbol_count = len(self._data)
                response_length = 120 * symbol_count
                response_value = struct.pack("II", symbol_count, response_length)

            elif index_group == constants.ADSIGRP_SYM_UPLOAD:
                response_value = b""
                for (group, offset) in self._data.keys():
                    response_value += struct.pack("III", 120, group, offset)
                    response_value += b"\x00" * 108

            else:
                # Create response of repeated 0x0F with a null
                # terminator for strings
                var = self.get_variable_by_indices(index_group, index_offset)
                response_value = var.value[:plc_datatype]

            return struct.pack("<I", len(response_value)) + response_value

        def handle_write() -> bytes:
            """Handle write request."""
            data = request.ams_header.data

            index_group = struct.unpack("<I", data[:4])[0]
            index_offset = struct.unpack("<I", data[4:8])[0]
            plc_datatype = struct.unpack("<I", data[8:12])[0]
            value = data[12 : (12 + plc_datatype)]

            logger.info(
                (
                    "Command received: WRITE (index group={}, index offset={}, "
                    "data length={}, value={}"
                ).format(hex(index_group), hex(index_offset), plc_datatype, value)
            )

            if index_group == constants.ADSIGRP_SYM_RELEASEHND:
                return b""

            elif index_group == constants.ADSIGRP_SYM_VALBYHND:
                var = self.get_variable_by_handle(index_offset)
                var.write(value, request)
                return b""

            var = self.get_variable_by_indices(index_group, index_offset)

            var.write(value, request)

            # no return value needed
            return b""

        def handle_read_write() -> bytes:
            """Handle read-write request."""
            data = request.ams_header.data

            # parse the request
            index_group = struct.unpack("<I", data[:4])[0]
            index_offset = struct.unpack("<I", data[4:8])[0]
            read_length = struct.unpack("<I", data[8:12])[0]
            write_length = struct.unpack("<I", data[12:16])[0]
            write_data = data[16 : (16 + write_length)]

            logger.info(
                (
                    "Command received: READWRITE "
                    "(index group={}, index offset={}, read length={}, "
                    "write length={}, write data={})"
                ).format(
                    hex(index_group),
                    hex(index_offset),
                    read_length,
                    write_length,
                    write_data,
                )
            )

            # Get variable handle by name if demanded
            if index_group == constants.ADSIGRP_SYM_HNDBYNAME:

                var_name = write_data.decode()

                # This could be part of a write-by-name, so create the
                # variable if it does not yet exist
                var = self.get_variable_by_name(var_name)

                read_data = struct.pack("<I", var.handle)

            # Get the symbol if requested
            elif index_group == constants.ADSIGRP_SYM_INFOBYNAMEEX:

                var_name = write_data.decode()
                var = self.get_variable_by_name(var_name)

                read_data = var.get_packed_info()

            # Write to a list of variables
            elif index_group == constants.ADSIGRP_SUMUP_WRITE:
                num_requests = index_offset  # number of requests is coded in the offset for sumup_write
                rq_list = [
                    (
                        struct.unpack("<I", write_data[i : i + 4])[0],  # index_group
                        struct.unpack("<I", write_data[i + 4 : i + 8])[
                            0
                        ],  # index_offset
                        struct.unpack("<I", write_data[i + 8 : i + 12])[0],  # size
                    )
                    for i in range(0, num_requests * 12, 12)
                ]

                data = write_data[num_requests * 12 :]
                offset = 0

                for index_group, index_offset, size in rq_list:
                    var = self.get_variable_by_indices(index_group, index_offset)
                    var.write(data[offset : offset + size], request)
                    offset += size

                read_data = struct.pack("<" + num_requests * "I", *(num_requests * [0]))

            # Read a list of variables
            elif index_group == constants.ADSIGRP_SUMUP_READ:
                num_requests = index_offset
                rq_list = [
                    (
                        struct.unpack("<I", write_data[i : i + 4])[0],  # index_group
                        struct.unpack("<I", write_data[i + 4 : i + 8])[
                            0
                        ],  # index_offset
                        struct.unpack("<I", write_data[i + 8 : i + 12])[0],  # size
                    )
                    for i in range(0, num_requests * 12, 12)
                ]

                read_data = struct.pack("<" + num_requests * "I", *(num_requests * [0]))
                for index_group, index_offset, size in rq_list:
                    var = self.get_variable_by_indices(index_group, index_offset)
                    read_data += var.value

            # Else just return the value stored
            else:

                # read stored data
                var = self.get_variable_by_indices(index_group, index_offset)
                read_data = var.value[:read_length]

                # store write data
                var.write(write_data, request)

            return struct.pack("<I", len(read_data)) + read_data

        def handle_read_state() -> bytes:
            """Handle read-state request."""
            logger.info("Command received: READ_STATE")
            ads_state = struct.pack("<H", constants.ADSSTATE_RUN)
            # I don't know what an appropriate value for device state is.
            # I suspect it may be unused..
            device_state = struct.pack("<H", 0)
            return ads_state + device_state

        def handle_writectrl() -> bytes:
            """Handle writectrl request."""
            logger.info("Command received: WRITE_CONTROL")
            # No response data required
            return b""

        def handle_add_devicenote() -> bytes:
            """Handle add_devicenode request.

            The actual callback is stored in `pyads_ex.callback_store`. All we need to do
            here is remember to prompt the client with an updated value if a callback was
            placed. The client will remember which callback belongs to it.
            """

            data = request.ams_header.data

            index_group, index_offset, length, mode, max_delay, cycle_time = \
                struct.unpack("<IIIIII", data[:24])

            logger.info(
                "Command received: ADD_DEVICE_NOTIFICATION (index_group={}, "
                "index_group={})".format(index_group, index_offset)
            )

            # Return value is the notification_handle
            # The notification handle is an incrementing value

            var = self.get_variable_by_indices(index_group, index_offset)

            handle = var.register_notification()

            return handle.to_bytes(4, byteorder='little')

        def handle_delete_devicenote() -> bytes:
            """Handle delete_devicenode request."""

            data = request.ams_header.data

            handle = struct.unpack("<I", data[:4])[0]

            logger.info("Command received: DELETE_DEVICE_NOTIFICATION (handle={})".format(handle))

            var = self.get_variable_by_notification_handle(handle)
            var.unregister_notification(handle)

            # No response data required
            return b""

        def handle_devicenote() -> bytes:
            """Handle a device notification."""
            logger.info("Command received: DEVICE_NOTIFICATION")
            # No response data required
            return b""

        # Function map
        function_map = {
            constants.ADSCOMMAND_READDEVICEINFO: handle_read_device_info,
            constants.ADSCOMMAND_READ: handle_read,
            constants.ADSCOMMAND_WRITE: handle_write,
            constants.ADSCOMMAND_READWRITE: handle_read_write,
            constants.ADSCOMMAND_READSTATE: handle_read_state,
            constants.ADSCOMMAND_WRITECTRL: handle_writectrl,
            constants.ADSCOMMAND_ADDDEVICENOTE: handle_add_devicenote,
            constants.ADSCOMMAND_DELDEVICENOTE: handle_delete_devicenote,
            constants.ADSCOMMAND_DEVICENOTE: handle_devicenote,
        }

        # Try to map the command id to a function, else return error code
        if command_id in function_map:
            content = function_map[command_id]()

        else:
            logger.info("Unknown Command: {0}".format(hex(command_id)))
            # Set error code to 'unknown command ID'
            error_code = "\x08\x00\x00\x00".encode("utf-8")
            return AmsResponseData(state, error_code, "".encode("utf-8"))

        # Set no error in response
        error_code = ("\x00" * 4).encode("utf-8")
        response_data = error_code + content

        return AmsResponseData(state, request.ams_header.error_code, response_data)

    def get_variable_by_handle(self, handle: int) -> PLCVariable:
        """Get PLC variable by handle, throw error when not found"""
        for idx, var in self._data.items():
            if var.handle == handle:
                return var

        raise KeyError(
            "Variable with handle `{}` not found - Create it first "
            "explicitly or write to it".format(handle)
        )

    def get_variable_by_indices(
        self, index_group: int, index_offset: int
    ) -> PLCVariable:
        """Get PLC variable by handle, throw error when not found"""
        tup = (index_group, index_offset)
        if tup in self._data:
            return self._data[tup]
        raise KeyError(
            "Variable with indices ({}, {}) not found - Create "
            "it first explicitly or write to it".format(index_group, index_offset)
        )

    def get_variable_by_name(self, name: str) -> PLCVariable:
        """Get variable by name, throw error if not found"""
        name = name.strip("\x00")
        for key, var in self._data.items():
            if var.name == name:
                return var
        raise KeyError(
            "Variable with name `{}` not found - Create it first "
            "explicitly or write to it".format(name)
        )

    def get_variable_by_notification_handle(self, handle: int) -> PLCVariable:
        """Get variable by a notification handle, throw error if not found"""
        for _, var in self._data.items():
            if handle in var.notifications:
                return var

        raise KeyError("Notification handle `{}` could not be resolved".format(handle))

    def add_variable(self, var: PLCVariable) -> None:
        """Add a new variable."""
        tup = (var.index_group, var.index_offset)
        self._data[tup] = var



import os
import struct
import subprocess
import sys
from pathlib import Path

from setuptools import setup
from setuptools.command.bdist_wheel import bdist_wheel, get_platform
from setuptools.command.build_py import build_py
from setuptools.command.install import install


def _is_32bit_interpreter() -> bool:
    return struct.calcsize("P") == 4


src_folder = Path(__file__).parent.absolute() / "src"
# ^ This will be on PATH for editable install
adslib_folder = Path(__file__).parent.absolute() / "adslib/build"
adslib_file = src_folder / "adslib.so"


class CustomBuildPy(build_py):
    """Custom command for `build_py`.

    This command class is used because it is always run, also for an editable install.
    """

    @classmethod
    def compile_adslib(cls) -> bool:
        """Return `True` if adslib was actually compiled."""
        if cls.platform_is_unix():
            cls._clean_library()
            cls._compile_library()
            return True

        return False

    @staticmethod
    def _compile_library():
        """Use `make` to build adslib - build is done in-place."""
        # Produce `adslib.so`:
        subprocess.call(["make", "-C", "adslib"])

    @staticmethod
    def _clean_library():
        """Remove all compilation artifacts."""
        patterns = (
            "*.a",
            "**/*.o",
            "*.bin",
            "*.so",
        )
        for pattern in patterns:
            for file in adslib_folder.glob(pattern):
                os.remove(file)

        if adslib_file.is_file():
            os.remove(adslib_file)

    @staticmethod
    def platform_is_unix():
        return sys.platform.startswith("linux") or sys.platform.startswith("darwin")

    def run(self):
        if self.compile_adslib():
            # Move .so file from Git submodule into src/ to have it on PATH:
            self.move_file(
                str(adslib_folder / "adslib.so"),
                str(adslib_file),
            )

        super().run()


class CustomInstall(install):
    """Install compiled adslib (but only for Linux)."""

    def run(self):
        if CustomBuildPy.platform_is_unix():
            adslib_dest = Path(self.install_lib)
            if not adslib_dest.exists():
                adslib_dest.mkdir(parents=True)
            self.copy_file(
                str(adslib_file),
                str(adslib_dest),
            )
        super().run()


class CustomBDistWheel(bdist_wheel):
    """Manually mark our wheel for a specific platform."""

    def get_tag(self):
        """

        See https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/
        """
        impl_tag = "py3"  # Same wheel across Python versions
        abi_tag = "none"  # Same wheel across ABI versions (not a C-extension)

        # But we need to differentiate on the platform for the compiled adslib

        # The following code is copied from `setuptools.command.bdist_wheel.get_tag()`
        plat_name = get_platform(self.bdist_dir)

        if _is_32bit_interpreter():
            if plat_name in ("linux-x86_64", "linux_x86_64"):
                plat_name = "linux_i686"
            if plat_name in ("linux-aarch64", "linux_aarch64"):
                plat_name = "linux_armv7l"

        plat_name = (
            plat_name.lower().replace("-", "_").replace(".", "_").replace(" ", "_")
        )

        return impl_tag, abi_tag, plat_name


# noinspection PyTypeChecker
setup(
    cmdclass={
        "build_py": CustomBuildPy,
        "install": CustomInstall,
        "bdist_wheel": CustomBDistWheel,
    },
)
